/*
 **	canvas.cpp
 **	by Saki Sakira <sakira@sun.dhis.portside.net>
 **/

#include <stdio.h>
#include <qvbox.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qdirectpainter_qws.h>
#include <qarray.h>
#include <qimage.h>
#include <qmenubar.h>
#include <qstring.h>
#include <qpopupmenu.h>
#include <qfileinfo.h>
#include <qmessagebox.h>
//#include <qtopia/qpeapplication.h>
#include <qpe/qpeapplication.h>
#include <qdatetime.h>
#include <qtimer.h>
#include <qfile.h>
#include <qdir.h>

#include "constants.h"
#include "spainter.h"
#include "colordialog.h"
#include "layeredpainter.h"
#include "toolpanel.h"
#include "tspressure.h"
#include "layerpanel.h"
#include "colorpanel.h"
#include "ptptformat.h"
//#include "filedialog.h"
#include "canvas.h"
#include "clipboard.h"
#include "z_color_adjust.h"

////////// Canvas

uint16* Canvas::_col_adj_table;

Canvas::Canvas(QWidget *pparent)
     : QWidget(pparent)
{
  parent = (SPainter*)pparent;
  
  setEnabled(TRUE);
  setMouseTracking(FALSE);
  setFocusPolicy(QWidget::ClickFocus);

  img_modified = true;
  fullscreen = TRUE;
  pick_color_mode = FALSE;
  zoom_mode_flag = FALSE;
  shift_mode = false;
  layer_shift_mode = false;
  select_mode = SelectModeNone;
  pen_mode = 0;
  pen_width_diff = 0;
  layer_mode = LayerModeNone;
  zoom = shift_x = shift_y = 0;
  loupe_zoom = 1;
  loupe_x = loupe_y = 0;
  opt_panel_type = None;
  panel_position = TopTools;
  pressing_tool = 0;
  timer = new QTimer(this);

  sp = 0;
  pix = 0;
  tools = 0;
  x0 = x1 = y0 = y1 = -1;

  pencil = PtPen(black, 4);
  brush = PtPen(red, 10);
  eraser = PtPen(white, 10);
  cloud = PtPen(lightGray, 10);
  pen = &pencil;

  pen_density = 128;
  last_pen_method = SolidBrush;
  antialias = true;
  dropping_mouse_release = true;
  apply_press_width = apply_press_density = false;

  layers.setNumOfLayers(1);

  pts.setAutoDelete(TRUE);

  repaint();
}

Canvas::~Canvas() {
#if 0
  delete sp;
  delete pix;
#endif
}

void Canvas::set_orientation(void) {
  int last_o = orientation;
  QDirectPainter dp(this);

  orientation = dp.transformOrientation();
  xoff = dp.xOffset();
  yoff = dp.yOffset();

  if (sp) {
    sp->setOrientation(orientation);
    if (last_o != orientation && panel_position) {
      update_info_rect(true);
    }
  }
}

void Canvas::adjustColorTable(bool f) {
  _col_adj_table = _z_col_adjust.adjust_table(f);
}

void Canvas::setupPainter(void) {
  int w, h;

  _col_adj_table = _z_col_adjust.adjust_table();

  parent->setFullScreen();
  QWidget *d = QPEApplication::desktop();
  w = d->width();
  h = d->height();

  set_orientation();

  int loupe_width = ((umin(w, h) / 2) >> ZoomMax) << ZoomMax;
  loupe_size.setWidth(loupe_width);
  loupe_size.setHeight(loupe_width);

  if (orientation) {
    dwidth = h;
    dheight = w;
    tools = new ToolPanel(height());
  } else {
    dwidth = w;
    dheight = h;
    tools = new ToolPanel(width());
  }
  parent->setTools(tools);
  parent->setColorPanel(&colorpanel);

  TSPressure::setWidth(dwidth);
  TSPressure::setHeight(dheight);

  sp = new LayeredPainter(dwidth, dheight);
  sp->setToolPanel(tools);
  sp->setShowingTools(true);

  sp->fill();
  sp->setLayer(LayeredPainter::BaseLayer);
//  tools = sp->getTools();
  tools->setActive(ToolPanel::Pencil, true);

  set_pen_width(pen->width());
  set_pen();
  set_antialias(true);

//  QImage img;
  pix = new QPixmap();
  sp->getQImage(&img, rect(), orientation);
  pix->convertFromImage(img, 0);
  img.detach();

  connect(sp, SIGNAL(toolPanelChanged()),
          this, SLOT(update_info_rect()));
  connect(sp, SIGNAL(showingChanged(int, bool)),
          &layers, SLOT(setShowing(int, bool)));
  connect(sp, SIGNAL(compositionChanged(int, int)),
          &layers, SLOT(setComposition(int, int)));
  connect(sp, SIGNAL(alphaChanged(int, int)),
          &layers, SLOT(setAlpha(int, int)));
  connect(sp, SIGNAL(currentChanged(int)),
          &layers, SLOT(setCurrent(int)));
  connect(sp, SIGNAL(currentChanged(int)),
	  this, SLOT(layerChanged(int)));
  connect(sp, SIGNAL(numOfLayersChanged(int)),
          &layers, SLOT(setNumOfLayers(int)));
  connect(&colorpanel, SIGNAL(selected(uint16)),
          this, SLOT(getColor(uint16)));
  connect(&colorpanel, SIGNAL(finished(void)),
          this, SLOT(finishColorPanel(void)));
  connect(&colorpanel, SIGNAL(showDialog(QColor)),
          this, SLOT(set_pen_color_dlg(QColor)));
  connect(&colorpanel, SIGNAL(paletteChanged(void)),
          this, SLOT(setUserPalette(void)));
  connect(timer, SIGNAL(timeout(void)),
          this, SLOT(tapping(void)));
  
  clearImage(false);

  // add by anoda 2003.10.27
  parent->ldx();
}

void Canvas::convert_geometry(int* x, int* y) {
  if (!zoom) return;

  if (zoom > 0) {
    if (orientation) {
      *x = shift_y + (*x >> zoom);
      *y = (dwidth - 1 - shift_x - (dwidth >> zoom)) + (*y >> zoom);
    } else {
      *x = shift_x + (*x >> zoom);
      *y = shift_y + (*y >> zoom);
    }
  } else {
    int unzoom = -zoom;
    if (orientation) {
      *x = shift_y + (*x << unzoom);
      *y = (dwidth - 1 - shift_x - (dwidth << unzoom)) + (*y << unzoom);
    } else {
      *x = shift_x + (*x << unzoom);
      *y = shift_y + (*y << unzoom);
    }
  }
}

void Canvas::convert_loupe_geometry(int *x, int *y) {
  int last_shift_x = shift_x;
  int last_shift_y = shift_y;
  int last_zoom = zoom;

  shift_x = loupe_x;
  shift_y = loupe_y;
  zoom = loupe_zoom;
  convert_geometry(x, y);
  shift_x = last_shift_x;
  shift_y = last_shift_y;
  zoom = last_zoom;
}

void Canvas::disconvert_geometry(int* x, int* y) {
  if (!zoom) return;

  if (zoom > 0) {
    if (orientation) {
      *x = (*x - shift_y) << zoom;
      *y = (*y  - (dwidth - 1 - shift_x - (dwidth >> zoom))) << zoom;
    } else {
      *x = (*x - shift_x) << zoom;
      *y = (*y - shift_y) << zoom;
    }
  } else {
    int unzoom = -zoom;
    if (orientation) {
      *x = (*x - shift_y) >> unzoom;
      *y = (*y - (dwidth - 1 - shift_x - (dwidth << unzoom))) >> unzoom;
    } else {
      *x = (*x - shift_x) >> unzoom;
      *y = (*y - shift_y) >> unzoom;
    }
  }
}

QPoint Canvas::convert_geometry(const QPoint &p) {
  int x = p.x();
  int y = p.y();

  convert_geometry(&x, &y);

  return QPoint(x, y);
}

void Canvas::update_rect(const QRect &rect) {
  if (!rect.width() || !rect.height()) return;

  if (zoom > 0)
    update_rect_zoom(rect);
  else if (zoom < 0)
    update_rect_unzoom(rect);
  else
    update_rect_nozoom(rect);
}

void Canvas::update_rect_nozoom(const QRect &rect) {
  if (!fullscreen) return;
  
  unsigned long int *buf, *fbuf;
  int x, y, w2, fw2, w, h;
  int x0, x1, y0, y1, yc;
  int offset, foffset, soffset;
  int sx, sy;

  w = sp->get_width();
  h = sp->get_height();
  x0 = rect.left() >> 1;
  x1 = rect.right() >> 1;
  y0 = rect.top();
  y1 = rect.bottom();
  	// copy for 4 bytes (= 2 pixels) each

  buf = (unsigned long int*)sp->frameBuffer();
  QDirectPainter dp(this);
  fbuf = (unsigned long int*)dp.frameBuffer();
  fw2 = dp.lineStep() >> 2;
  w2 = w >> 1;

  //  sx = (w - shift_x) % w;
  //  sy = (h - shift_y) % h;
  sx = shift_x >> 1;
  sy = shift_y;
  while (sx < 0) sx += w;
  while (sy < 0) sy += h;
  sx %= w;
  sy %= h;

  soffset = sy * w2 + sx;
  if (sy + y1 < h - 1)
    yc = y1;
  else
    yc = h - 2 - sy;

  for (y = y0; y < yc; y ++) {
    foffset = (y * fw2) + x0;
    offset = (y * w2) + x0 + soffset;
    for (x = x0; x <= x1; x ++)
      fbuf[foffset ++] = adjust(buf[offset ++]);
  }

  if (yc <= y1) {
    int xc;
    if (sx + x1 < w - 1)
      xc = x1;
    else
      xc = w2 - 2 - sx;
    foffset = (yc * fw2) + x0;
    offset = (yc * w2) + x0 + soffset;
    for (x = x0; x <= xc; x ++)
      fbuf[foffset ++] = adjust(buf[offset ++]);
    offset = 0;
    for (x = xc + 1; x <= x1; x ++)
      fbuf[foffset ++] = adjust(buf[offset ++]);
  }

  soffset = (- yc - 1) * w2 + sx;
  for (y = yc + 1; y <= y1; y ++) {
    foffset = (y * fw2) + x0;
    offset = (y * w2) + x0 + soffset;
    for (x = x0; x <= x1; x ++)
      fbuf[foffset ++] = adjust(buf[offset ++]);
  }
}

void Canvas::update_rect_zoom(const QRect &rect) {
  /*
   ** assume that 2**zoom divides dp.lineStep()
   ** and that zoom > 0.
   */
  
  if (!zoom) {
    update_rect_nozoom(rect);
    return;
  } else if (zoom < 0) {
    update_rect_unzoom(rect);
    return;
  }

  if (!fullscreen) return;
  
  unsigned long int *fbuf;
  int xoff, yoff, w, h, fw2;
  int x, y, xx, yy;
  int x0, x1, y0, y1;
  int offset, foffset, foffset2;

  uint16 *sp_buf = sp->frameBuffer();

  w = sp->get_width();
  h = sp->get_height();
  x0 = rect.left() >> zoom;
  x1 = rect.right() >> zoom;
  y0 = rect.top() >> zoom;
  y1 = rect.bottom() >> zoom;
  if (x1 & ((1 << zoom) - 1)) x1 ++;
  if (y1 & ((1 << zoom) - 1)) y1 ++;

  if ((x0 << zoom) >= dwidth) 
    x0 = x1 = (dwidth - 1) >> zoom;
  else if ((x1 << zoom) >= dwidth) 
    x1 = (dwidth - 1) >> zoom;

  if ((y0 << zoom) >= dheight) 
    y0 = y1 = (dheight - 1) >> zoom;
  else if ((y1 << zoom) >= dheight) 
    y1 = (dheight - 1) >> zoom;

  QDirectPainter dp(this);
  fbuf = (unsigned long int*)dp.frameBuffer();
  fw2 = dp.lineStep() >> 2;

  if (orientation) {
    xoff = dp.yOffset();
    yoff = dp.xOffset();
  } else {
    xoff = dp.xOffset();
    yoff = dp.yOffset();
  }

  xoff = yoff = 0; // debug

  int grain = (1 << zoom);
  int grain2 = (grain >> 1);
  int wh = w * h;
  int offset_x, shift_x_x;
  unsigned long int pix;
  offset = (shift_y + y0) * w - xoff + shift_x;
  foffset = (yoff + (y0 << zoom)) * fw2 - (xoff >> 1);
  for (y = y0; y <= y1; y ++) {
    for (x = x0; x <= x1; x ++) {
      offset_x = offset + x;
      shift_x_x = shift_x + x;
      if (offset_x >= 0 && offset_x < wh
          && shift_x_x >= 0 &&  shift_x_x < w) {
	pix = adjust(sp_buf[offset_x]);
	pix = pix | (pix << 16);
      } else 
	pix = 0x0000ffff;
      foffset2 = foffset + (x << (zoom - 1));
      for (yy = 0; yy < grain; yy ++) {
        for (xx = 0; xx < grain2; xx ++)
          fbuf[foffset2 + xx] = pix;

        foffset2 += fw2;
      }
    }
    foffset += fw2 << zoom;
    offset += w;
  }
}

void Canvas::update_rect_unzoom(const QRect &rect) {
  if (zoom > 0) {
    update_rect_zoom(rect);
    return;
  } else if (zoom == 0) {
    update_rect_nozoom(rect);
    return;
  }
  
  if (!fullscreen) return;

  static const uint16 gray_color = 0x8410;
  int unzoom = - zoom;
  uint32 *sbuf = (uint32*)sp->frameBuffer();
  uint16 *fbuf;
  int w, h, fw, s_x, s_y;
  int x, y, xx, yy, x_unzoom, yc;
  int x0, x1, y0, y1, x1r, y1r;
  int soffset, foffset, soffset2, s_offset;
  uint32 r, g, b;

  w = sp->get_width();
  h = sp->get_height();
  x0 = rect.left();
  y0 = rect.top();
  x1 = rect.right();
  y1 = rect.bottom();
  x1r = min(x1, (w - 1) >> unzoom);
  y1r = min(y1, (h - 1) >> unzoom);

  int w_half = w >> 1;
  s_x = shift_x >> unzoom + 1;
  s_y = shift_y >> unzoom;
  while (s_x < 0) s_x += w;
  while (s_y < 0) s_y += h;
  s_x %= w >> unzoom;
  s_y %= h >> unzoom;
  s_offset = s_y * w_half + s_x << unzoom;
  if (s_y + y1r < (h - 1 >> unzoom))
    yc = y1r;
  else
    yc = ((h - 1) >> unzoom) - 1 - s_y;
  
  QDirectPainter dp(this);
  fbuf = (uint16*)dp.frameBuffer();
  fw = dp.lineStep() >> 1;
  int grain = (1 << unzoom);
  int grain2 = grain >> 1;

  for (y = y0; y < yc; y ++) {
    soffset = (y << unzoom) * w_half + s_offset;
    foffset = y * fw;
    for (x = x0; x <= x1r; x ++) {
      x_unzoom = x << (unzoom - 1);
      r = g = b = 0;
      for (yy = 0; yy < grain; yy ++) {
        soffset2 = soffset + yy * w_half + x_unzoom;
        for (xx = 0; xx < grain2; xx ++)
          add_each_color(sbuf[soffset2 ++], r, g, b);
      }
      fbuf[foffset + x] = adjust(pack_each_color(r, g, b));
    }
    for ( ; x <= x1; x ++)
      fbuf[foffset + x] = gray_color;
  }

  if (yc <= y1) {
    int xc;
    if (s_x + x1r < (w - 1) >> unzoom)
      xc = x1r;
    else
      xc = ((w - 1) >> unzoom) - 1 - s_x;
    y = yc;
    soffset = (y << unzoom) * w_half + s_offset;
    foffset = y * fw;
    for (x = x0; x <= xc; x ++) {
      x_unzoom = x << (unzoom - 1);
      r = g = b = 0;
      for (yy = 0; yy < grain; yy ++) {
        soffset2 = soffset + yy * w_half + x_unzoom;
        for (xx = 0; xx < grain2; xx ++) 
          add_each_color(sbuf[soffset2 ++], r, g, b);
      }
      fbuf[foffset + x] = adjust(pack_each_color(r, g, b));
    }
    soffset = 0;
    for ( ; x <= x1r; x ++) {
      x_unzoom = x << (unzoom - 1);
      r = g = b = 0;
      for (yy = 0; yy < grain; yy ++) {
        soffset2 = soffset + yy * w_half + x_unzoom;
        for (xx = 0; xx < grain2; xx ++)
          add_each_color(sbuf[soffset2 ++], r, g, b);
      }
      fbuf[foffset + x] = adjust(pack_each_color(r, g, b));
    }
    for ( ; x <= x1; x ++)
      fbuf[foffset + x] = gray_color;
  }

  s_offset = (- yc - 1) * w_half + s_x << unzoom;
  for (y = yc + 1; y <= y1r; y ++) {
    soffset = (y << unzoom) * w_half + s_offset;
    foffset = y * fw;
    for (x = x0 ; x <= x1r; x ++) {
      x_unzoom = x << (unzoom - 1);
      r = g = b = 0;
      for (yy = 0; yy < grain; yy ++) {
        soffset2 = soffset + yy * w_half + x_unzoom;
        for (xx = 0; xx < grain2; xx ++)
          add_each_color(sbuf[soffset2 ++], r, g, b);
      }
      fbuf[foffset + x] = adjust(pack_each_color(r, g, b));
    }
    for ( ; x <= x1; x ++)
      fbuf[foffset + x] = gray_color;
  }

  for ( ; y <= y1; y ++) {
    foffset = y * fw;
    for (x = x0; x <= x1; x ++)
      fbuf[foffset + x] = gray_color;
  }
}

void Canvas::update_panel_rect(const QSize &size,
                               uint16 *buf16) {
  if (!panel_position || !fullscreen)
    return;
  
  int x0, y0;
  int x, y, l, ti, i, w2, fw2;
  unsigned long int *buf, *fbuf;

  if (panel_position == TopTools) {
    x0 = dwidth - size.width();
    y0 = 0;
  } else {
    x0 = 0;
    y0 = dheight - size.height();
  }

  buf = (uint32*)buf16;
  QDirectPainter dp(this);
  fbuf = (unsigned long int*)dp.frameBuffer();

  w2 = size.width() >> 1;
  fw2 = dp.lineStep() >> 2;
  l = w2 * size.height();
  for (y = 0; y < size.height(); y ++) {
    ti = w2 * y;
    i = (y0 + y) * fw2 + (x0 >> 1);
    for (x = 0; x < w2; x ++)
      fbuf[i + x] = adjust(buf[ti + x]);
  }
}

void Canvas::update_optional_panel_rect(const QSize &size,
                                        uint16 *buf16) {
  if (!panel_position || !fullscreen) return;

  int x0, y0;
  int x, y, l, ti, i, w2, fw2;
  unsigned long int *buf, *fbuf;

  if (panel_position == TopTools) {
    x0 = dwidth - size.width();
    y0 = info_rect.height();
  } else {
    x0 = 0;
    y0 = dheight - info_rect.height() - size.height();
  }

  buf = (uint32*)buf16;
  QDirectPainter dp(this);
  fbuf = (unsigned long int*)dp.frameBuffer();

  w2 = size.width() >> 1;
  fw2 = dp.lineStep() >> 2;
  l = w2 * size.height();
  for (y = 0; y < size.height(); y ++) {
    ti = w2 * y;
    i = (y0 + y) * fw2 + (x0 >> 1);
    for (x = 0; x < w2; x ++)
      fbuf[i + x] = adjust(buf[ti + x]);
  }
}

void Canvas::update_tools_rect(void) {
  update_panel_rect(tools->size(), tools->infoImage());
}

void Canvas::update_colorpanel_rect(void) {
  if (!panel_position) return;

  update_optional_panel_rect(colorpanel.size(),
                    colorpanel.infoImage());
  
  QPoint hpt = opt_panel_rect.topLeft() +
    colorpanel.huePoint(orientation);
  QPoint slpt = opt_panel_rect.topLeft() +
    colorpanel.satlumPoint(orientation);
  const int r = 5;

  QPainter p(this);
  p.setPen(PtPen(Qt::white, 2));
  p.setRasterOp(XorROP);

  if (orientation)
    p.drawLine(hpt.x(), hpt.y(), hpt.x() + 9, hpt.y());
  else
    p.drawLine(hpt.x(), hpt.y(), hpt.x(), hpt.y() + 9);
  
  p.drawEllipse(slpt.x() - r/2, slpt.y() - r/2, r, r);
}

void Canvas::update_layers_rect(void) {
  if (opt_panel_type != Layers) return;

  QPoint pt;
  if (orientation) {
    if (panel_position == TopTools)
      pt = tools_rect.topRight() + QPoint(1, 0);
    else
      pt = tools_rect.topLeft() + QPoint(-layers.width(), 0);
  } else {
    if (panel_position == TopTools)
      pt = tools_rect.bottomRight() + QPoint(-layers.width(), 1);
    else
      pt = tools_rect.topRight() +
        QPoint(-layers.width(), -layers.height());
  }

  layers.setTopLeft(pt);
  bitBlt(this, pt, layers.pixmap(), layers.pixmap()->rect());
}

void Canvas::update_loupe_rect(void) {
  update_loupe_rect(opt_panel_rect);
}

void Canvas::update_loupe_rect(QRect r) {
  if (!panel_position) return;

  int last_zoom = zoom;
  int last_shift_x = shift_x;
  int last_shift_y = shift_y;

  zoom = loupe_zoom;

  if (orientation) {
    shift_x = loupe_x;
    shift_y = loupe_y;
    QRect or_(dwidth - r.y() - r.height(), r.x(),
             r.height(), r.width());
    update_rect_zoom(or_);
  } else {
    shift_x = loupe_x;
    shift_y = loupe_y;
    update_rect_zoom(r);
  }
  
  zoom = last_zoom;
  shift_x = last_shift_x;
  shift_y = last_shift_y;
}

void Canvas::update_optional_panel(void) {
  if (!panel_position || !fullscreen) return;

  switch (opt_panel_type) {
  case Layers:
    update_layers_rect();
    break;
  case Color:
    update_colorpanel_rect();
    break;
  case Loupe:
    update_loupe_rect();
    break;
  default:
    break;
  }
}

void Canvas::update_panels(void) {
  if (!fullscreen) return;

  update_tools_rect();
  update_optional_panel();
}

void Canvas::update_image_info(void) {
  int x;
  QStringList info;

  sp->infoStrings(info);
  
  QPainter p(this);
  p.setBackgroundColor(Qt::white);
  p.setBackgroundMode(Qt::OpaqueMode);
    // it seems not working on SL-C760...

  QRect rect = p.boundingRect(0, 0, width(), height(),
                              AlignLeft,
                              "* S 255 normal");
  x = width() - 20 - rect.width();
  
  if (panel_position == TopTools) {
    p.setPen(Qt::white);
    for (uint i = 0; i < info.count(); i ++)
      p.drawText(x, 30 + i*20, info[info.count() - 1 - i]);

    p.setPen(Qt::black);
    for (uint i = 0; i < info.count(); i ++)
      p.drawText(x + 1, 31 + i*20, info[info.count() - 1 - i]);
  } else if (panel_position == BottomTools) {
    p.setPen(Qt::white);
    for (uint i = 0; i < info.count(); i ++)
      p.drawText(x,
                 height() - 30 - (info.count() - i) * 20,
                 info[info.count() - 1 - i]);

    p.setPen(Qt::black);
    for (uint i = 0; i < info.count(); i ++)
      p.drawText(x + 1,
                 height() - 29 - (info.count() - i) * 20,
                 info[info.count() - 1 - i]);
  }
}

void Canvas::updateRectR0(const QRect &rect) {
  update_rect(rect);
}

void Canvas::updateRectR270(const QRect &rect) {
  int x, y, w, h;

  x = dwidth - 1 - rect.bottom();
  y = rect.left();
  w = rect.height();
  h = rect.width();

  update_rect(QRect(x, y, w, h));
}

void Canvas::updateRect(const QRect &rect) {
  if (orientation)
    updateRectR270(rect);
  else
    updateRectR0(rect);
}

void Canvas::paintEvent(QPaintEvent *e) {
  if (!sp) setupPainter();

  set_orientation();
  QArray<QRect> rects = e->region().rects();
  QRect *r;

  if (fullscreen) {
    for (uint i = 0; i < rects.count(); i ++)
      updateRect(rects[i]);

    update_panels();
  } else {
    if (img_modified) {
      sp->getQImage(&img, qApp->desktop()->rect(), orientation);
      pix->convertFromImage(img);
      img.detach();
      img_modified = false;
    }

    for (uint i = 0; i < rects.count(); i ++) {
      r = &rects[(int)i];
      bitBlt(this, r->left(), r->top(), pix, 
	     r->left() + xoff , r->top() + yoff, r->width(), r->height());
    }

    update_image_info();
  }
}

void Canvas::mousePressEvent(QMouseEvent *e) {
  if (!fullscreen) return;

  if (select_mode == SelectModeCopy) {
    select_copy_press_event(e);
    return;
  } else if (select_mode == SelectModePaste) {
    //    if (fullscreen && panel_position && tools_rect.contains(e->pos()))
    if (panel_position && tools_rect.contains(e->pos()))
      panel_event(e);
    else
      select_paste_press_event(e);
    return;
  }

  timer->start(press_delay, true);
  TSPressure::clear();

  //  set_orientation();

  //  if (fullscreen && panel_position) {
  if (panel_position) {
    if (tools_rect.contains(e->pos())) {
      panel_event(e);
      return;
    } else if (opt_panel_type &&
               opt_panel_rect.contains(e->pos())) {
      if (opt_panel_type == Layers) 
        layerpanel_event(e);
      else if (opt_panel_type == Color)
        colorpanel_event(e);
      else if (opt_panel_type == Loupe)
        loupe_event(e);

      return;
    }
  } 

  int x = e->globalX();
  int y = e->globalY();
  convert_geometry(&x, &y);
  
  if (pick_color_mode) {
    pick_color(e);
    return;
  } else if (zoom_mode_flag) {
    set_zoom(x, y);
    return;
  } else if (opt_panel_type == Loupe) {
    set_loupe_by_center(x, y);
    return;
  }

  sp->backupCurrentLayer();
  
  if (e->state() & Qt::ShiftButton) {
    bool last_dropping_mouse_release =
      dropping_mouse_release;
    dropping_mouse_release = false;

    if (x0 < 0) {
      px1 = x1 = e->globalX();
      py1 = y1 = e->globalY();
      convert_geometry(&x1, &y1);
      if (orientation) {
        int x = x1;
        x1 = dwidth - 1 - y1;
        y1 = x;
      }

      px0 = px1;
      py0 = py1;
      x0 = x1;
      y0 = y1;
    }

    mouseMoveEvent(e);

    dropping_mouse_release =
      last_dropping_mouse_release;
    return;
  }

  px1 = x1 = e->globalX();
  py1 = y1 = e->globalY();
  convert_geometry(&x1, &y1);
  if (orientation) {
    int x = x1;
    x1 = dwidth - 1 - y1;
    y1 = x;
  }

#if 0
  if (x1 < 0) x1 = 0;
  if (y1 < 0) y1 = 0;
  if (x1 >= dwidth) x1 = dwidth - 1;
  if (y1 >= dheight) y1 = dheight - 1;
#endif
  x1 = max(0, min(x1, sp->get_width() - 1));
  y1 = max(0, min(y1, sp->get_height() - 1));

  px0 = px1;
  py0 = py1;
  x0 = x1;
  y0 = y1;

  if (shift_mode)
    return;

  pts.clear();

  hideInfo();

  mouseMoveEvent(e);
}

void Canvas::mouseMoveEvent(QMouseEvent *e) {
  if (!fullscreen) return;

  if (select_mode == SelectModePre &&
      !tools_rect.contains(e->pos())) {
    set_antialias_mode();
    set_select_mode();
    panel_move_event(e);
    return;
  } else if (select_mode == SelectModeCopy) {
    select_copy_move_event(e);
    return;
  } else if (select_mode == SelectModePaste) {
    select_paste_move_event(e);
    return;
  } else if (select_mode) return;

  if ((e->state() & Qt::ShiftButton) &&
      dropping_mouse_release) {
    return;
  }
  
  if (pen_mode && panel_position &&
      !tools_rect.contains(e->pos())) {
    pressing_tool = 0;
    panel_move_event(e);
    return;
  } else if (layer_mode && panel_position &&
             opt_panel_type == Layers &&
             !layers.rect().contains(e->pos())) {
    pressing_tool = 0;
    layerpanel_move_event(e);
    return;
  } else if (opt_panel_rect.contains(e->pos()) &&
             opt_panel_type == Color && x0 < 0) {
    pressing_tool = 0;
    colorpanel_move_event(e);
    return;
  } else if (opt_panel_rect.contains(e->pos()) &&
             opt_panel_type == Loupe && x0 >= 0) {
    loupe_move_event(e);
    return;
  }

  if (opt_panel_type == Loupe && panel_position) x0 = -1;
  if (x0 < 0) return;
  pressing_tool = 0;

  PtPen ppen(*pen);
  int nzx, nzy;

  x0 = x1; y0 = y1;
  px0 = px1; py0 = py1;
  
  nzx = px1 = x1 = e->globalX();
  nzy = py1 = y1 = e->globalY();
  convert_geometry(&x1, &y1);
  if (orientation) {
    int x = x1;
    x1 = dwidth - 1 - y1;
    y1 = x;
    x = nzx;
    nzx = dwidth - 1 - nzy;
    nzy = x;
  }

  if (apply_press_width || apply_press_density) {
    //    int pressure = TSPressure::pressure(); // assume 0..256
    int pressure = TSPressure::pressure(nzx, nzy);

    if (pressure >= 0) {
      if (apply_press_width) {
        int w = ppen.width() * pressure * 3 / 2;
        if (w <= 0) w = 1;
        if (w > (PenWidthMax << 8)) w = PenWidthMax << 8;
	w |= 1;
        ppen.setDWidth(w);
      }

      if (apply_press_density) {
        int d = (pen_density * pressure) >> 8;
        sp->setPenDensity(d);
      }
    } else {
      sp->setPenDensity(pen_density);
      set_pen_pressure_width(false);
      set_pen_pressure_density(false);
    }
  }
  
  if (dropping_mouse_release &&
      (px0 - px1) * (px0 - px1)
      + (py0 - py1) * (py0 - py1) > 10000) {
    px0 = px1; py0 = py1;
    x0 = x1; y0 = y1;
  }

  if (shift_mode) {
    shift_image_by_mouse();
    if (x1 < 0) x1 = 0;
    if (y1 < 0) y1 = 0;
    if (x1 >= dwidth) x1 = dwidth - 1;
    if (y1 >= dheight) y1 = dheight - 1;
    x0 = x1; y0 = y1;
    return;
  } else if (layer_shift_mode) {
    layer_shift_image_by_mouse();
    if (x1 < 0) x1 = 0;
    if (y1 < 0) y1 = 0;
    if (x1 >= dwidth) x1 = dwidth - 1;
    if (y1 >= dheight) y1 = dheight - 1;
    x0 = x1; y0 = y1;
    return;
  }

#if 0
  if (x1 < 0) x1 = 0;
  if (y1 < 0) y1 = 0;
  if (x1 >= dwidth) x1 = dwidth - 1;
  if (y1 >= dheight) y1 = dheight - 1;
#endif
  x1 = max(0, min(x1, sp->get_width() - 1));
  y1 = max(0, min(y1, sp->get_height() - 1));

  sp->setOrientation(orientation);
  sp->setPen(ppen, false);
  sp->drawLine(x0, y0, x1, y1, true);

  int min_px, max_px, min_py, max_py;
  int pwz;
  if (zoom >= 0)
    pwz = ((pen->width() << zoom) >> 1) + 1;
  else
    pwz = ((pen->width() >> -zoom) >> 1) + 1;
  if (px0 <= px1) {
    min_px = px0 - pwz;
    max_px = px1 + pwz;
  } else {
    min_px = px1 - pwz;
    max_px = px0 + pwz;
  }
  if (py0 <= py1) {
    min_py = py0 - pwz;
    max_py = py1 + pwz;
  } else {
    min_py = py1 - pwz;
    max_py = py0 + pwz;
  }

  updateRect(QRect(min_px, min_py,
		   max_px - min_px + 1,
		   max_py - min_py + 1) & rect());
  img_modified = true;
}

void Canvas::mouseReleaseEvent(QMouseEvent *e) {
  if (!fullscreen) return;

  sp->setPen(*pen, true, true);
  sp->setPenDensity(pen_density, true, true);

  if (select_mode) {
    panel_release_event(e);
    update_panels();
    return;
  }

  if (layer_shift_mode) {
    sp->updateRect(QRect(0, 0, dwidth, dheight));
    updateRect(QRect(0, 0, width(), height()));
  }
  
  pressing_tool = 0;
  TSPressure::clear();

  info_string.truncate(0);

  if (pen_mode) {
    if (tools_rect.contains(e->pos())) {
      switch (pen_mode) {
      case PenModePressureWidth:
	if (parent->noviceMode() && sp->numOfLayers() == 2 &&
	    sp->current() == 1)
	  sp->setLayer(0);

        set_pencil();
        break;
      case PenModeDensity:
	if (parent->noviceMode() && sp->numOfLayers() == 2 &&
	    sp->current() == 0)
	  sp->setLayer(1);

        set_brush();
        break;
      case PenModePressureDensity:
        set_eraser();
        break;
      case PenModeCloud:
        set_cloud();
        break;
      case PenModeColor:
        set_pick_mode();
        break;
      case PenModeMove:
	set_shift_mode();
	break;
      case PenModeAntialias:
	tools->setActive(ToolPanel::Select, false);
	break;
      case PenModeZoom:
        {
          int x, y;

          if (orientation) {
            x = tools_rect.height() - 1 - e->globalY() + tools_rect.y();
            y = e->globalX() - tools_rect.x();
          } else {
            x = e->globalX() - tools_rect.x();
            y = e->globalY() - tools_rect.y();
          }
          if (tools->isLeftside(x, y))
            set_unzoom();
          else
            zoom_mode();
        }
        break;
      }
      pen_mode = 0;
      pen_width_diff = 0;
    } else {
      if (pen_mode == PenModeMove) {
	set_layer_shift_mode();
	pen_mode = 0;
	pen_width_diff = 0;
      } else {
	pen_mode = 0;
	pen_width_diff = 0;
	if (pen == &pencil)
	  set_pencil();
	else if (pen == &brush)
	  set_brush();
	else if (pen == &eraser)
	  set_eraser();
	else if (pen == &cloud)
	  set_cloud();
      }
    }

    tools->clearSelectionHalf();
    set_pen_pressure_width(apply_press_width);
    set_pen_pressure_density(apply_press_density);
    update_info_string();

    return;
  } else if (layer_mode) {
    layerpanel_release_event(e);
    return;
  }

  if (opt_panel_type == Color &&
      panel_position &&
      opt_panel_rect.contains(e->pos()) && x0 < 0) {
    colorpanel_release_event(e);
    return;
  } else if (opt_panel_type == Loupe &&
             panel_position) {
    loupe_release_event(e);
    return;
  }

  if (e->state() & Qt::ShiftButton) return;
  
  x0 = x1 = y0 = y1 = -1;
  px0 = px1 = py0 = py1 = -1;

  sp->clearMask();
  sp->setPen(*pen);
  
  parent->setFullScreen();
  update_panels();

//  set_orientation();
  
  return;
}

void Canvas::keyPressEvent(QKeyEvent *e) {
  bool last_fullscreen;

  switch (e->key()) {
  case Qt::Key_Escape:
    quit_application();
    break;
  case Qt::Key_Return:
  case Qt::Key_F33:	// OK key
  case 0x2010:		// "*" beside LCD
  case Qt::Key_T:
    show_hide_tools();
    break;
  case Qt::Key_Down:
    change_pen_width(-1);
    break;
  case Qt::Key_Up:
    change_pen_width(+1);
    break;
  case Qt::Key_Left:
    switch (leftright) {
    case LeftRightSwitchMethod:
      set_pen_method(get_pen_method() + 1);
      break;
    case LeftRightDensity:
      set_pen_density(pen_density - 8);
      set_brush();
      break;
    }
    break;
  case Qt::Key_Right:
    switch (leftright) {
    case LeftRightSwitchMethod:
      set_pen_method(get_pen_method() + NumOfPens - 1);
      break;
    case LeftRightDensity:
      set_pen_density(pen_density + 8);
      set_brush();
      break;
    }
    break;
  case Qt::Key_K:
    sp->switchLayer();
    repaint();
    break;
  case Qt::Key_E:
    set_eraser();
    break;
  case Qt::Key_W:
    set_brush();
    break;
  case Qt::Key_B:
    set_pencil();
    break;
  case Qt::Key_R:
    set_cloud();
    break;
  case Qt::Key_C:
    set_color_panel();
//    set_color();
    break;
  case Qt::Key_D:
    joinShowingLayers();
    break;
  case Qt::Key_M:
    show_hide_menu();
    break;
  case Qt::Key_P:
    set_pick_mode();
    break;
  case Qt::Key_Z:
    if (e->text() == QString("z"))
      sp->backupCurrentLayer();
    else {
      sp->restoreCurrentLayer();
      img_modified = true;
      repaint();
    }
    break;
  case Qt::Key_X:
    sp->exchangeCurrentLayer();
    img_modified = true;
    repaint();
    break;
  case Qt::Key_L:
    if (e->text() == QString("l"))
      zoom_mode();
    else
      set_unzoom();
    break;
  case Qt::Key_1:
  case Qt::Key_2:
  case Qt::Key_3:
  case Qt::Key_4:
  case Qt::Key_5:
  case Qt::Key_6:
  case Qt::Key_7:
  case Qt::Key_8:
  case Qt::Key_9:
    set_pen_density(((e->key() - Qt::Key_1) +1) * 25);
    break;
  case Qt::Key_A:
    set_antialias(!antialias);
    break;
  case Qt::Key_Space:
    set_shift_mode();
    break;
  case Qt::Key_V:
    img_modified = true;
    exchangeColorPaperPen();
    break;
  case Qt::Key_H:
    sp->setShowing(!sp->getShowing());
    img_modified = true;
    repaint();
    break;
  case Qt::Key_S:
    last_fullscreen = fullscreen;
    fullscreen = false;

    if (e->text() == QString("s")) {
      repaint();
      parent->save();
    } else
      parent->saveImage();

    parent->enableFullscreen();
    fullscreen = last_fullscreen;
    break;
  default:
    e->ignore();
  }
}

void Canvas::keyReleaseEvent(QKeyEvent *e) {
  if (!(e->state() & Qt::ShiftButton)) {
    x0 = -1;
  }
  e->ignore();
}

void Canvas::panel_event(QMouseEvent *e) {
  int x, y;

  x0 = PanelEventIndex;

  if (orientation) {
//    x = panel_rect.height() - 1 - e->globalY() - panel_rect.y();
    x = tools_rect.height() - 1 - e->globalY() + tools_rect.y();
    y = e->globalX() - tools_rect.x();
  } else {
    x = e->globalX() - tools_rect.x();
    y = e->globalY() - tools_rect.y();
  }
  
  pen_width_diff = tools->isLeftside(x, y) ? -1 : 1;
    
    switch (tools->getIndex(x, y)) {
    case ToolPanel::Undo:
      pen_width_diff = 0;
      sp->exchangeCurrentLayer();
      img_modified = true;
      repaint();
      break;
    case ToolPanel::Menu:
      pen_width_diff = 0;
      show_hide_menu();
      break;
    case ToolPanel::Select:
      pen_width_diff = 0;
      set_select_mode();
      break;
    case ToolPanel::Move:
      pen_width_diff = 0;
      set_move_mode();
      break;
    case ToolPanel::Zoom:
      pen_width_diff = 0;
      set_zoom_mode();
      break;
    case ToolPanel::Cloud:
      set_cloud_mode(tools->isLeftside(x, y));
      previous_value = sp->cloudDensity();
      break;
    case ToolPanel::Eraser:
      set_eraser_mode(tools->isLeftside(x, y));
      break;
    case ToolPanel::Brush:
      set_brush_mode(tools->isLeftside(x, y));
      previous_value = pen_density;
      break;
    case ToolPanel::Pencil:
      set_pencil_mode(tools->isLeftside(x, y));
      break;
    case ToolPanel::PenInfo:
      pen_width_diff = 0;
      set_peninfo_mode();
      break;
    case ToolPanel::Layer:
      pen_width_diff = 0;
      show_hide_layers();
      break;
    case ToolPanel::Palette:
      pen_width_diff = 0;
      set_pen();
      pen->setColor(tools->getColor(x, y));
      set_pen();
      break;
    }

    if (tools->getIndex(x, y) != ToolPanel::Move) {
      unset_layer_shift_mode();
      unset_shift_mode();
    }
}

void Canvas::colorpanel_event(QMouseEvent *e) {
  int x, y;

  x0 = ColorPanelEventIndex;

  if (orientation) {
    x = opt_panel_rect.height() - 1 - e->globalY() + opt_panel_rect.y();
    y = e->globalX() - opt_panel_rect.x();
  } else {
    x = e->globalX() - opt_panel_rect.x();
    y = e->globalY() - opt_panel_rect.y();
  }

  colorpanel.clicked(x, y, false);
  update_optional_panel();
}

void Canvas::layerpanel_event(QMouseEvent *e) {
  int x, y, c;
  uint l;

  x0 = LayerPanelEventIndex;

  x = e->globalX() - opt_panel_rect.x();
  y = e->globalY() - opt_panel_rect.y();

  layers.editing = l = layers.layerNum(y);

  switch (layers.typeNum(x)) {
  case LayerPanel::Show:
    sp->setShowing(l, !sp->getShowing(l));
    img_modified = true;
    repaint();
    break;
  case LayerPanel::Composition:
    c = (layers.composition(l) + 1) % NumOfComposition;
    sp->setCompositionMethod(l, c);
    img_modified = true;
    repaint();
    break;
  case LayerPanel::Alpha:
    layer_mode = LayerModeAlpha;
    break;
  case LayerPanel::Tab:
    layer_mode = LayerModeTab;
    break;
  }
}

void Canvas::loupe_event(QMouseEvent *e) {
  sp->backupCurrentLayer();

  px1 = x1 = e->globalX();
  py1 = y1 = e->globalY();
  convert_loupe_geometry(&x1, &y1);
  if (orientation) {
    int x = x1;
    x1 = dwidth - 1 - y1;
    y1 = x;
  }

  px0 = px1;
  py0 = py1;
  x0 = x1;
  y0 = y1;

  pts.clear();

  loupe_move_event(e);
}

void Canvas::panel_move_event(QMouseEvent *e) {
  int x, y;

  if (orientation) {
    x = tools_rect.height() - 1 - e->globalY() + tools_rect.y();
    y = e->globalX() - tools_rect.x();
  } else {
    x = e->globalX() - tools_rect.x();
    y = e->globalY() - tools_rect.y();
  }

    tools->clearSelectionHalf();
    int diff;
    
    switch (pen_mode) {
    case PenModeDensity:
      tools->setActive(ToolPanel::Brush, true);
      set_pen_density(300 * x / dwidth - 10);
      info_string = QString::number(pen_density) + "\n";
      diff = (int)pen_density - previous_value;
      if (diff >= 0) info_string += "+";
      info_string += QString::number(diff);
      update_tools_rect();
      update_info_string(true);
      break;
    case PenModeCloud:
      tools->setActive(ToolPanel::Cloud, true);
      set_cloud_density(300 * x / dwidth - 10);
      info_string = QString::number(sp->cloudDensity()) + "\n";
      diff = (int)sp->cloudDensity() - previous_value;
      if (diff >= 0) info_string += "+";
      info_string += QString::number(diff);
      update_tools_rect();
      update_info_string(true);
      break;
    case PenModePressureWidth:
      tools->setActive(ToolPanel::Pencil, true);
      set_pen_pressure_width(x < dwidth / 2);
      break;
    case PenModePressureDensity:
      tools->setActive(ToolPanel::Eraser, true);
      set_pen_pressure_density(x < dwidth / 2);
      break;
    case PenModeAntialias:
      tools->setActive(ToolPanel::Select, true);
      set_antialias(x < dwidth / 2);
      break;
    case PenModeColor:
      set_color_panel();
      tools->setActive(ToolPanel::PenInfo, false);
      pen_mode = 0;
      break;
    case PenModeZoom:
      set_loupe_panel();
      tools->setActive(ToolPanel::Zoom, false);
      pen_mode = 0;
      break;
    }
}

void Canvas::colorpanel_move_event(QMouseEvent *e) {
  if (x0 != ColorPanelEventIndex) return;

  int x, y;

  if (orientation) {
    x = opt_panel_rect.height() - 1 - e->globalY() + opt_panel_rect.y();
    y = e->globalX() - opt_panel_rect.x();
  } else {
    x = e->globalX() - opt_panel_rect.x();
    y = e->globalY() - opt_panel_rect.y();
  }

  colorpanel.clicked(x, y, true);
  update_optional_panel();
}

void Canvas::layerpanel_move_event(QMouseEvent *e) {
  int x;

  if (orientation) {
    x = dwidth - 1 - e->globalY();
  } else {
    x = e->globalX();
  }

  if (layer_mode == LayerModeAlpha) {
    set_layer_alpha(layers.editing, 300 * x / dwidth - 10);
  }
}

void Canvas::loupe_move_event(QMouseEvent *e) {
  if (x0 < 0) return;
  pressing_tool = 0;

  PtPen ppen(*pen);
  int nzx, nzy;

  x0 = x1; y0 = y1;
  px0 = px1; py0 = py1;
  
  nzx = px1 = x1 = e->globalX();
  nzy = py1 = y1 = e->globalY();
  convert_loupe_geometry(&x1, &y1);

  int bx0, by0, bx1, by1;
  if (orientation) {
    int x = x1;
    x1 = dwidth - 1 - y1;
    y1 = x;
    x = nzx;
    nzx = dwidth - 1 - nzy;
    nzy = x;
  }
  bx0 = x0; by0 = y0; bx1 = x1; by1 = y1;

  sp->setOrientation(orientation);
  sp->setPen(ppen, false);
  sp->drawLine(x0, y0, x1, y1, true);

  int min_bx, max_bx, min_by, max_by;
  int pw;
  if (zoom >= 0)
    pw = ((pen->width() << zoom) >> 1) + 1;
  else
    pw = ((pen->width() >> -zoom) >> 1) + 1;
  disconvert_geometry(&bx0, &by0);
  disconvert_geometry(&bx1, &by1);
  if (bx0 <= bx1) {
    min_bx = bx0 - pw;
    max_bx = bx1 + pw;
  } else {
    min_bx = bx1 - pw;
    max_bx = bx0 + pw;
  }
  if (by0 <= by1) {
    min_by = by0 - pw;
    max_by = by1 + pw;
  } else {
    min_by = by1 - pw;
    max_by = by0 + pw;
  }
  update_rect(QRect(min_bx, min_by,
                   max_bx - min_bx + 1,
                   max_by - min_by + 1) & rect());

  int min_px, max_px, min_py, max_py;
  int pwz = ((pen->width() << loupe_zoom) >> 1) + 1;
  if (px0 <= px1) {
    min_px = px0 - pwz;
    max_px = px1 + pwz;
  } else {
    min_px = px1 - pwz;
    max_px = px0 + pwz;
  }
  if (py0 <= py1) {
    min_py = py0 - pwz;
    max_py = py1 + pwz;
  } else {
    min_py = py1 - pwz;
    max_py = py0 + pwz;
  }
  update_loupe_rect(QRect(min_px, min_py,
                          max_px - min_px + 1,
                          max_py - min_py + 1) & opt_panel_rect);
  img_modified = true;
}

void Canvas::panel_release_event(QMouseEvent *e) {
  int x, y;

  if (orientation) {
    x = tools_rect.height() - 1 - e->globalY() + tools_rect.y();
    y = e->globalX() - tools_rect.x();
  } else {
    x = e->globalX() - tools_rect.x();
    y = e->globalY() - tools_rect.y();
  }

  if (select_mode == SelectModePre) {
    select_mode = SelectModeCopy;
    return;
  } else if (select_mode == SelectModeCopy) {
    select_copy_release_event(e);
    return;
  } else if (select_mode) return;
}

void Canvas::colorpanel_release_event(QMouseEvent *e) {
  int x, y;

  if (orientation) {
//    x = panel_rect.height() - 1 - e->globalY() - panel_rect.y();
    x = opt_panel_rect.height() - 1 - e->globalY() + opt_panel_rect.y();
    y = e->globalX() - opt_panel_rect.x();
  } else {
    x = e->globalX() - opt_panel_rect.x();
    y = e->globalY() - opt_panel_rect.y();
  }

  colorpanel.released(x, y);
}

void Canvas::layerpanel_release_event(QMouseEvent *e) {
 int y;
  uint l;

  y = e->globalY() - layers.y();
  l = layers.layerNum(y);

  if (opt_panel_type == Layers && layers.rect().contains(e->pos())) {
    if (layers.editing == l)
      sp->setLayer(l);
    else if (layer_mode == LayerModeTab) {
      exchangeLayers(layers.editing, l);
    }
  } else
   fix_layer_alpha();

  layer_mode = LayerModeNone;
}

void Canvas::loupe_release_event(QMouseEvent *) {
  x0 = x1 = y0 = y1 = -1;
  px0 = px1 = py0 = py1 = -1;

  sp->clearMask();
  sp->setPen(*pen);
  
  parent->setFullScreen();
  update_panels();
}

void Canvas::joinShowingLayers(void) {
  bool last_fullscreen;

  last_fullscreen = fullscreen;
  fullscreen = false;

  if (!QMessageBox::warning(this, "PetitePeinture",
                         "Join all showing layers?\n\n",
                         "OK", "Cancel"))
    sp->joinShowingLayers();

  fullscreen = last_fullscreen;

  eraser.setColor(sp->paperColor());
}

void Canvas::exchangeLayersBaseColor(void) {
  exchangeLayers(LayeredPainter::BaseLayer,
                 LayeredPainter::ColorLayer);
}

void Canvas::exchangeLayersBaseOver(void) {
  exchangeLayers(LayeredPainter::BaseLayer,
                 LayeredPainter::OverLayer);
}

void Canvas::exchangeLayersColorOver(void) {
  exchangeLayers(LayeredPainter::ColorLayer,
                 LayeredPainter::OverLayer);
}

void Canvas::exchangeLayers(int i0, int i1) {
  sp->exchangeLayers(i0, i1);

  img_modified = true;

  eraser.setColor(sp->paperColor());
}

void Canvas::moveUpCurrentLayer(void) {
  int c = sp->current();
  exchangeLayers(c, c + 1);
  //  sp->switchLayer(+1);
}

void Canvas::moveDownCurrentLayer(void) {
  int c = sp->current();
  exchangeLayers(c - 1, c);
  //  sp->switchLayer(-1);
}

void Canvas::addNewLayer(void) {
  sp->createLayer();
}

void Canvas::duplicateCurrentLayer(void) {
  sp->duplicateCurrentLayer();
}

void Canvas::deleteCurrentLayer(void) {
  if (QMessageBox::warning(this, "PetitePeinture",
                           "Destroy this Layer?\n\n",
                           "OK", "Cancel"))
    return;
  
  sp->deleteCurrentLayer();

  img_modified = true;
}

void Canvas::setMinComposition(void) {
  sp->setCompositionMethod(MinComposition);

  img_modified = true;
}

void Canvas::setMaxComposition(void) {
  sp->setCompositionMethod(MaxComposition);

  img_modified = true;
}

void Canvas::setMulComposition(void) {
  sp->setCompositionMethod(MulComposition);

  img_modified = true;
}

void Canvas::setScreenComposition(void) {
  sp->setCompositionMethod(ScreenComposition);

  img_modified = true;
}

void Canvas::setSatComposition(void) {
  sp->setCompositionMethod(SatComposition);

  img_modified = true;
}

void Canvas::setColComposition(void) {
  sp->setCompositionMethod(ColComposition);

  img_modified = true;
}

void Canvas::setNormalComposition(void) {
  sp->setCompositionMethod(NormalComposition);

  img_modified = true;
}

void Canvas::setMaskComposition(void) {
  sp->setCompositionMethod(MaskComposition);

  img_modified = true;
}

void Canvas::setAlphaChannelComposition(void) {
  sp->setCompositionMethod(AlphaChannelComposition);

  img_modified = true;
}

void Canvas::update_info_string(bool bg) {
  if (info_string.isEmpty()) return;
  
  QRect rect = tools_rect;
  rect.setWidth(rect.width() - 1);
  rect.setHeight(rect.height() - 1);
  QPainter p(this);

  p.setPen(Qt::white);
  if (bg) {
    QBrush b(Qt::white);
    QRect inside = p.boundingRect(rect, Qt::AlignCenter,
                                  info_string);
    p.fillRect(inside, b);
  } else {
    p.drawText(rect, Qt::AlignCenter, info_string);
  }
  
  rect.moveBy(1, 1);
  p.setPen(Qt::black);
  p.drawText(rect, Qt::AlignCenter, info_string);
}

void Canvas::update_info_rect(void) {
  update_info_rect(false);
}

void Canvas::update_info_rect(bool) {
  QRect rect;
  QPoint p = mapToGlobal(QPoint(0, 0));

  rect = info_rect;
  rect.moveBy(-p.y(), -p.x());
  update_rect(rect);

  if (panel_position) {
    QSize size = tools->size();
    QSize osize;
    int x0, y0, ox0, oy0;

    switch (opt_panel_type) {
    case Layers:
      osize = layers.size();
      break;
    case Color:
      osize = colorpanel.size();
      break;
    case Loupe:
      osize = loupe_size;
    default:
      break;
    }
    
    if (panel_position == TopTools) {
      x0 = dwidth - size.width();
      y0 = 0;
      ox0 = dwidth - osize.width();
      oy0 = size.height();
    } else {
      x0 = 0;
      y0 = dheight - size.height();
      ox0 = (opt_panel_type == Layers) ? dwidth - osize.width() : 0;
      oy0 = y0 - osize.height();
      if (orientation && opt_panel_type == Layers) 
	oy0 = y0 - osize.width();
    }

    if (orientation) {
      tools_rect = QRect(y0, dwidth - size.width() - x0,
                         size.height(), size.width());
      if (opt_panel_type == Layers)
	opt_panel_rect = QRect(oy0, dwidth - osize.width() - ox0,
			       osize.width(), osize.height());
      else
	opt_panel_rect = QRect(oy0, dwidth - osize.width() - ox0,
			       osize.height(), osize.width());
    } else {
      tools_rect = QRect(x0, y0,
                         size.width(), size.height());
      opt_panel_rect = QRect(ox0, oy0,
                             osize.width(), osize.height());
    }

    info_rect = QRect(x0, y0,
                      size.width(), size.height());

    update_panels();
    update_info_string();

    return;
  }
}

void Canvas::set_antialias(bool a) {
  sp->setAntialias(a);
  antialias = a;
}

void Canvas::setSyncWidth(bool f) {
  sync_width = f;
}

void Canvas::setLeftRight(uint i) {
  leftright = i;
}

void Canvas::setSeparatePen(bool s) {
  separate_pen = s;
}

uint Canvas::get_pen_method(void) {
  if (pen == &pencil)
    return PenPencil;
  else if (pen == &brush)
    return PenWater;
  else if (pen == &eraser)
    return PenEraser;
  else
    return PenCloud;
}

void Canvas::set_pen_method(uint m) {
  m = m % NumOfPens;
  pen_width_diff = 0;
  
  switch (m) {
  case PenPencil:
    set_pencil();
    break;
  case PenWater:
    set_brush();
    break;
  case PenEraser:
    set_eraser();
    break;
  case PenCloud:
    set_cloud();
    break;
  }
}

void Canvas::set_pen_width(int w) {
  if (x0 >= 0) return;
  if (w < 1 || w > PenWidthMax) return;

  if (sync_width && pen != &eraser) {
    pencil.setWidth(w);
    brush.setWidth(w);
    //    eraser.setWidth(w);  // DO comment-out this
    cloud.setWidth(w);
  } else {
    pen->setWidth(w);
  }

  sp->setPen(*pen);
  info_rect = sp->showPenInfo();
  info_string.setNum(w);
  update_info_rect();
  info_string.truncate(0);

  emit pencilWidthChanged(pencil.width());
  emit brushWidthChanged(brush.width());
  emit eraserWidthChanged(eraser.width());
  emit cloudWidthChanged(cloud.width());
}

void Canvas::set_pen_density(int d) {
  if (x0 >= 0) return;

  if (d < 0)
    d = 0;
  else if (d > 256)
    d = 256;
  
  sp->setPenDensity(d);
  pen_density = d;

  info_string.setNum(pen_density);
  
  info_rect = sp->showPenInfo();
  update_info_rect();
}

void Canvas::set_cloud_density(int d) {
  if (x0 >= 0) return;

  if (d < 0)
    d = 0;
  else if (d > 255)
    d = 255;

  
  sp->setCloudDensity(d);
  sp->setPenMethod(NumOfBrushes);

  info_string.setNum(d);

  info_rect = sp->showPenInfo();
  update_info_rect();
}

void Canvas::change_pen_width(int diff) {
  set_pen_width(pen->width() + diff);

  pen_width_diff = 0;
}

void Canvas::change_pen_width(void) {
  change_pen_width(pen_width_diff);
}

void Canvas::set_eraser_mode(bool dec) {
  if (x0 >= 0) return;

  pressing_tool = dec ? ToolPanel::EraserL : ToolPanel::EraserR;
  
  if (pen != &eraser)
    pen_width_diff = 0;

  tools->clearSelectionHalf();
  if (dec) {
    tools->setActive(ToolPanel::EraserL, true);
  } else {
    tools->setActive(ToolPanel::EraserR, true);
  }

  pen_mode = PenModePressureDensity;
}

void Canvas::set_pen(void) {
  switch (last_pen_method) {
  case SolidBrush:
  case EraserBrush:
    set_pencil();
    break;
  case WaterBrush:
    set_brush();
    break;
  }
}

void Canvas::set_pencil(void) {
  if (x0 >= 0) return;

  pen = &pencil;
  change_pen_width();

  tools->clearSelection();
  tools->setActive(ToolPanel::Pencil, true);

  sp->setPenMethod(SolidBrush);
  sp->setPen(*pen);
  info_rect = sp->showPenInfo();
  info_string.setNum(pen->width());
  update_info_rect();

  last_pen_method = SolidBrush;
}

void Canvas::set_brush(void) {
  if (x0 >= 0) return;

  pen = &brush;
  change_pen_width();

  tools->clearSelection();
  tools->setActive(ToolPanel::Brush, true);

  sp->setPenMethod(WaterBrush);
  sp->setPen(*pen);
  info_rect = sp->showPenInfo();
  info_string = QString::number(pen->width()) + "\n";
  info_string += QString::number(pen_density);
  update_info_rect();

  last_pen_method = WaterBrush;
}

void Canvas::set_eraser(void) {
  if (x0 >= 0) return;

  eraser.setColor(sp->paperColor());
  pen = &eraser;
  change_pen_width();

  tools->clearSelection();
  tools->setActive(ToolPanel::Eraser, true);

  //  sp->setPenMethod(SolidBrush);
  sp->setPenMethod(EraserBrush);
  sp->setPen(*pen);
  info_rect = sp->showPenInfo();
  info_string.setNum(pen->width());
  update_info_rect();
}

void Canvas::set_brush_mode(bool down) {
  if (x0 >= 0) return;

  pressing_tool = down ? ToolPanel::BrushL : ToolPanel::BrushR;
  
  if (pen != &brush)
    pen_width_diff = 0;

  set_brush();
  tools->clearSelection();
  
  tools->clearSelectionHalf();
  if (down) {
    tools->setActive(ToolPanel::BrushL, true);
  } else {
    tools->setActive(ToolPanel::BrushR, true);
  }

  pen_mode = PenModeDensity;
}

void Canvas::set_pencil_mode(bool down) {
  if (x0 >= 0) return;

  pressing_tool = down ? ToolPanel::PencilL : ToolPanel::PencilR;
  
  if (pen != &pencil)
    pen_width_diff = 0;

  tools->clearSelectionHalf();
  if (down) {
    tools->setActive(ToolPanel::PencilL, true);
  } else {
    tools->setActive(ToolPanel::PencilR, true);
  }

  pen_mode = PenModePressureWidth;
}

#if 0
void Canvas::set_cloud(void) {
  static int cloud_m = CloudWeakBrush;
  if (x0 >= 0) return;

  if (sp->penMethod() >= CloudWeakBrush &&
      sp->penMethod() <= CloudWideBrush) {
    switch (cloud_m) {
    case CloudWeakBrush:
      cloud_m = CloudMidBrush;
      break;
    case CloudMidBrush:
      cloud_m = CloudWideBrush;
      break;
    case CloudWideBrush:
    default:
      cloud_m = CloudWeakBrush;
    }
  }
  
  tools->clearSelectionHalf();
  switch (cloud_m) {
  case CloudWeakBrush:
    tools->setActive(ToolPanel::CloudL, true);
    break;
  case CloudMidBrush:
    tools->setActive(ToolPanel::CloudR, true);
    break;
  case CloudWideBrush:
    tools->setActive(ToolPanel::Select, true);
    break;
  }

  pen = &brush;

  sp->setPenMethod(cloud_m);
  sp->setPen(*pen);
  update_info_rect();
  info_string.truncate(0);
  
  cloud_density_mode = false;

  printf("edge:%d method:%d \n",
         antialias, cloud_m);
}
#endif

void Canvas::set_cloud(void) {
  if (x0 >= 0) return;

  pen = &cloud;
  change_pen_width();

  tools->clearSelection();
  tools->setActive(ToolPanel::Cloud, true);

  sp->setPenMethod(CloudWideBrush);
  sp->setPen(*pen);
  info_rect = sp->showPenInfo();
  info_string = QString::number(pen->width()) + "\n";
  info_string +=  QString::number(sp->cloudDensity());
  update_info_rect();
}

void Canvas::set_cloud_mode(bool down) {
  if (x0 >= 0) return;

  pressing_tool = down ? ToolPanel::CloudL : ToolPanel::CloudR;
  
  if (pen != &cloud)
    pen_width_diff = 0;

  set_cloud();
  tools->clearSelection();
  
  tools->clearSelectionHalf();
  if (down) {
    tools->setActive(ToolPanel::CloudL, true);
  } else {
    tools->setActive(ToolPanel::CloudR, true);
  }
  
  pen_mode = PenModeCloud;
}

void Canvas::set_antialias_mode(void) {
  if (x0 >= 0) return;

  tools->clearSelection();
  tools->setActive(ToolPanel::Select, true);

  pen_mode = PenModeAntialias;
}

void Canvas::set_layer_alpha(int alpha) {
  if (x0 >= 0) return;

  if (alpha < 0)
    alpha = 0;
  else if (alpha > 255)
    alpha = 255;
  
  sp->setLayerAlpha(alpha);

  updateRect(QRect(0, 0, width(), height()));

  info_string.setNum(alpha);
  update_info_string();
}

void Canvas::set_layer_alpha(int l, int alpha) {
  if (x0 >= 0) return;

  if (alpha < 0)
    alpha = 0;
  else if (alpha > 255)
    alpha = 255;
  
  sp->setLayerAlpha(l, alpha);

  updateRect(QRect(0, 0, width(), height()));

  info_string.setNum(alpha);
  update_info_string();
}

void Canvas::fix_layer_alpha(void) {
  sp->setLayer();

  update_info_rect();
  info_string.truncate(0);

  layer_mode = LayerModeNone;
}

void Canvas::set_color_panel(void) {
  opt_panel_type = Color;

  colorpanel.setInitColor(pen->color());
  
  update_info_rect();
  
  updateRect(QRect(0, 0, width(), height()));
  repaint();
}

void Canvas::set_loupe_panel(void) {
  if (opt_panel_type == Loupe) {
    if (++ loupe_zoom > ZoomMax) loupe_zoom = 1;
  } else {
    opt_panel_type = Loupe;
    loupe_x = loupe_y = 0;
  }

  update_info_rect();
  repaint();
}

void Canvas::finishColorPanel(void) {
  opt_panel_type = None;

  update_info_rect();

  updateRect(QRect(0, 0, width(), height()));
  repaint();
}

void Canvas::setUserPalette(void) {
  tools->setColorPalette();
  update_tools_rect();
}

void Canvas::getColor(uint16 col) {
  set_pen();
  pen->setColor(SketchPainter::unpack_color(col));
  set_pen();
}

void Canvas::set_pen_color_dlg(QColor col) {
  bool last_fullscreen;
//  QColor col = brush.color();
  ColorDialog d(col, "Pen Color");

  last_fullscreen = fullscreen;
  fullscreen = FALSE;
  
  if (d.exec()) {
    set_pen();
    pen->setColor(col);
    set_pen();
  }

  parent->enableFullscreen();
  
  opt_panel_type = None;
  fullscreen = last_fullscreen;
}

void Canvas::setPaperColor(void) {
  bool last_fullscreen;
  QColor col = sp->paperColor();
  ColorDialog d(col, "Paper Color");

  last_fullscreen = fullscreen;
  fullscreen = false;

  if (d.exec()) {
    sp->setPaperColor(col);
    img_modified = true;
  }

  fullscreen = last_fullscreen;
}

void Canvas::exchangeColorPaperPen(void) {
  QColor pa_col = sp->paperColor();
  QColor pe_col = brush.color();

  set_pen();
  pen->setColor(pa_col);
  sp->setPaperColor(pe_col);

  set_eraser();
  set_pen();

  repaint();
}

void Canvas::set_peninfo_mode(void) {
  pen_mode = PenModeColor;

  tools->setActive(ToolPanel::PenInfo, true);
  update_panels();
}

void Canvas::set_pick_mode(void) {
  pick_color_mode = TRUE;

  tools->setActive(ToolPanel::PenInfo, true);
  update_panels();

  hideInfo();
}

void Canvas::pick_color(QMouseEvent *e) {
  int px, py;

  QPoint p(convert_geometry(e->globalPos()));
  if (orientation) {
    px = dwidth - 1 - p.y();
    py = p.x();
  } else {
    px = p.x();
    py = p.y();
  }

  tools->setActive(ToolPanel::PenInfo, false);

  QColor pcol = sp->pickColor(px, py);
  set_pen();
  pen->setColor(pcol);
  colorpanel.setColor(pcol);
  set_pen();

  pick_color_mode = FALSE;
}

void Canvas::set_shift_by_center(int x, int y) {
  int px, py;

  if (orientation) {
    px = dwidth - 1 - y;
    py = x;
  } else {
    px = x;
    py = y;
  }

  if (zoom >= 0) {
    px -= dwidth >> (zoom + 1);
    py -= dheight >> (zoom + 1);
  } else {
    px -= dwidth << (-zoom - 1);
    py -= dheight << (-zoom - 1);
  }
  
  shift_x = px;
  shift_y = py;
  
  adjust_shift();
}

void Canvas::set_loupe_by_center(int x, int y) {
  int px, py;

  if (orientation) {
    px = dwidth - 1 - y;
    py = x;

    px -= loupe_size.height() >> (loupe_zoom + 1);
    py -= loupe_size.width() >> (loupe_zoom + 1);
 
    loupe_x = px - ((dwidth - opt_panel_rect.y()
                     - opt_panel_rect.height()) >> loupe_zoom);
    loupe_y = py - (opt_panel_rect.x() >> loupe_zoom);
  } else {
    px = x;
    py = y;

    px -= loupe_size.width() >> (loupe_zoom + 1);
    py -= loupe_size.height() >> (loupe_zoom + 1);
 
    loupe_x = px - (opt_panel_rect.x() >> loupe_zoom);
    loupe_y = py - (opt_panel_rect.y() >> loupe_zoom);
  }

  adjust_loupe();
}

void Canvas::adjust_shift(void) {
  if (zoom > 0) {
    int pw, ph, px, py;

    px = shift_x;
    py = shift_y;
    pw = sp->get_width();
    ph = sp->get_height();

    if (px < -pw + 1) px = -pw + 1;
    if (py < -ph + 1) py = -ph + 1;

    if (px > pw - 1) px = pw - 1;
    if (py > ph - 1) py = ph - 1;

    shift_x = px;
    shift_y = py;
  } else {
    int w = sp->get_width();
    int h = sp->get_height();

    shift_x = (shift_x + w) % w;
    shift_y = (shift_y + h) % h;
  }
}

void Canvas::adjust_loupe(void) {
  int px, py, pw, ph, px0, py0, pd;

  px = loupe_x;
  py = loupe_y;

  if (orientation) {
    pw = opt_panel_rect.height();
    ph = opt_panel_rect.width();
    px0 = dwidth - 1 - opt_panel_rect.y();
    py0 = opt_panel_rect.x();
  } else {
    ph = opt_panel_rect.height();
    pw = opt_panel_rect.width();
    px0 = opt_panel_rect.x();
    py0 = opt_panel_rect.y();
  }

  if (px < - (px0 >> loupe_zoom)) px = - px0 >> loupe_zoom;
  if (py < - (py0 >> loupe_zoom)) py = - py0 >> loupe_zoom;

  if ((pd = px + ((px0 + pw) >> loupe_zoom) - dwidth + 1) > 0)
    px -= pd;
  if ((pd = py + ((py0 + ph) >> loupe_zoom) - dheight + 1) > 0)
    py -= pd;

  loupe_x = px;
  loupe_y = py;
}

void Canvas::shift_image(int x, int y) {
  shift_x += x;
  shift_y += y;

  adjust_shift();

  update_rect(QRect(0, 0, dwidth, dheight));
}

void Canvas::shift_image_by_mouse(void) {
  int x, y;

  if (zoom >= 0) {
    if (orientation) {
      x = (py1 - py0) >> zoom;
      y = (px0 - px1) >> zoom;
    } else {
      x = (px0 - px1) >> zoom;
      y = (py0 - py1) >> zoom;
    }
  } else {
    int unzoom = -zoom;
    if (orientation) {
      x = (py1 - py0) << unzoom;
      y = (px0 - px1) << unzoom;
    } else {
      x = (px0 - px1) << unzoom;
      y = (py0 - py1) << unzoom;
    }
  }
  shift_image(x, y);
}

void Canvas::layer_shift_image(int dx, int dy) {
  QPoint pt(dx, dy);
  QRect rect(0, 0, width(), height());
  
  rect &= QRect(x1 - 100, y1 - 100, 200, 200);
  sp->shiftCurrentLayer(pt, rect);

  //  updateRect(QRect(0, 0, width(), height()));
  updateRect(rect);
}

void Canvas::layer_shift_image_by_mouse(void) {
  int x, y;

  if (zoom >= 0) {
    if (orientation) {
      x = (py1 - py0) >> zoom;
      y = (px0 - px1) >> zoom;
    } else {
      x = (px0 - px1) >> zoom;
      y = (py0 - py1) >> zoom;
    }
  } else {
    int unzoom = -zoom;
    if (orientation) {
      x = (py1 - py0) << unzoom;
      y = (px0 - px1) << unzoom;
    } else {
      x = (px0 - px1) << unzoom;
      y = (py0 - py1) << unzoom;
    }
  }
  layer_shift_image(x, y);
}

void Canvas::undo(void) {
    sp->exchangeCurrentLayer();
    img_modified = true;
    repaint();
}

void Canvas::set_select_mode(void) {
  if (select_mode == SelectModeNone) {
    select_mode = SelectModePre;
    sp->setToolPanelActive(ToolPanel::Select, true);
    sp->setToolPanelActive(ToolPanel::SelectL, false);
    sp->setToolPanelActive(ToolPanel::SelectR, false);
  } else if (select_mode == SelectModePaste) {
    PPClipboard *cb = sp->getClipboard();
    QPoint ppt(cb->x(), cb->y());

    sp->pasteClipboard(ppt);
    sp->setClipboardPasteMode(false);

    sp->updateRect(QRect(0, 0, dwidth, dheight));
    updateRect(QRect(0, 0, width(), height()));
    
    if (!fullscreen) {
      img_modified = true;
      repaint();
    }

    sp->setToolPanelActive(ToolPanel::Select, false);
    sp->setToolPanelActive(ToolPanel::SelectL, false);
    sp->setToolPanelActive(ToolPanel::SelectR, false);

    select_mode = SelectModeNone;
  } else {
    sp->setToolPanelActive(ToolPanel::Select, false);
    sp->setToolPanelActive(ToolPanel::SelectL, false);
    sp->setToolPanelActive(ToolPanel::SelectR, false);

    select_mode = SelectModeNone;
    sp->setClipboardPasteMode(false);

    sp->updateRect(QRect(0, 0, dwidth, dheight));
    updateRect(QRect(0, 0, width(), height()));
  }
}

void Canvas::set_select_mode_copy(void) {
  select_mode = SelectModeCopy;
  sp->setClipboardPasteMode(false);
}

void Canvas::set_select_mode_paste(void) {
  select_mode = SelectModePaste;
  sp->setClipboardPasteMode(true);
}

void Canvas::set_select_mode_paste_freeze(void) {
  if (select_mode == SelectModePaste)
    set_select_mode();
}

void Canvas::set_select_mode_paste_cancel(void) {
  if (select_mode == SelectModePaste)
    sp->setToolPanelActive(ToolPanel::Select, false);
    sp->setToolPanelActive(ToolPanel::SelectL, false);
    sp->setToolPanelActive(ToolPanel::SelectR, false);

    select_mode = SelectModeNone;
    sp->setClipboardPasteMode(false);

    sp->updateRect(QRect(0, 0, dwidth, dheight));
    updateRect(QRect(0, 0, width(), height()));
    
    if (!fullscreen) {
      img_modified = true;
      repaint();
    }
}

void Canvas::select_copy_press_event(QMouseEvent *e) {
  int x, y;

  pcopy_rect.setX(e->x());
  pcopy_rect.setY(e->y());
  
  x = e->globalX();
  y = e->globalY();

  convert_geometry(&x, &y);
  if (orientation) {
    int t = x;
    x = dwidth - 1 - y;
    y = t;
  }

  copy_rect.setX(x);
  copy_rect.setY(y);
  pcopy_rect.setWidth(0);

  sp->setToolPanelActive(ToolPanel::Select, false);
  sp->setToolPanelActive(ToolPanel::SelectL, true);
  sp->setToolPanelActive(ToolPanel::SelectR, false);
}

void Canvas::select_copy_move_event(QMouseEvent *e) {
  static int x, y;
  x = e->x();
  y = e->y();

  QPainter p(this);
  p.setPen(Qt::white);
  p.setRasterOp(Qt::XorROP);
  
  if (pcopy_rect.width())
    p.drawRect(pcopy_rect); // erase
  pcopy_rect.setWidth(0);

  pcopy_rect.setRight(x);
  pcopy_rect.setBottom(y);
  p.drawRect(pcopy_rect);
}

void Canvas::select_copy_release_event(QMouseEvent *e) {
  int x, y;
  x = e->globalX();
  y = e->globalY();
  convert_geometry(&x, &y);
  if (orientation) {
    int t = x;
    x = dwidth - 1 - y;
    y = t;
  }

  int x0, y0, x1, y1;
  x0 = min(x, copy_rect.x());
  x1 = max(x, copy_rect.x());
  y0 = min(y, copy_rect.y());
  y1 = max(y, copy_rect.y());

  copy_rect.setLeft(x0);
  copy_rect.setRight(x1);
  copy_rect.setTop(y0);
  copy_rect.setBottom(y1);

  sp->copyClipboard(copy_rect);

  sp->setToolPanelActive(ToolPanel::Select, false);
  sp->setToolPanelActive(ToolPanel::SelectL, false);
  sp->setToolPanelActive(ToolPanel::SelectR, true);

  if (fullscreen) {
    select_mode = SelectModePaste;
    sp->setClipboardPasteMode(true);
  } else
    select_mode = SelectModeNone;
}

void Canvas::select_paste_press_event(QMouseEvent *e) {
  QPainter p(this);
  p.setPen(Qt::white);
  p.setRasterOp(Qt::XorROP);
  
  if (pcopy_rect.width())
    p.drawRect(pcopy_rect); // erase
  pcopy_rect.setWidth(0);

  int x, y;
  x = e->globalX();
  y = e->globalY();
  convert_geometry(&x, &y);
  if (orientation) {
    int t = x;
    x = dwidth - 1 - y;
    y = t;
  }

  QPoint pt(x, y);

  if (sp->clipboardRect().contains(pt) || !fullscreen) {
    sp->setClipboardPasteMode(true);
    select_paste_move_event(e, false);
  } else {
    set_select_mode_paste_cancel();
  }
}

void Canvas::select_paste_move_event(QMouseEvent *e, bool move) {
  static int x0, x1, y0, y1;
  
  x0 = e->globalX();
  y0 = e->globalY();
  convert_geometry(&x0, &y0);
  if (orientation) {
    int t = x0;
    x0 = dwidth - 1 - y0;
    y0 = t;
  }

  if (!move) {
    x1 = x0;
    y1 = y0;
  }

  PPClipboard *cb = sp->getClipboard();
  QRect rect0 = cb->rect();
  cb->setX(cb->x() + (x0 - x1));
  cb->setY(cb->y() + (y0 - y1));
  QRect rect1 = cb->rect();

  sp->updateRect(rect0 & QRect(0, 0, dwidth, dheight));
  sp->updateRect(rect1 & QRect(0, 0, dwidth, dheight));
  
  bool last_fullscreen = fullscreen;
  fullscreen = true;

  update_rect(QRect(0, 0, dwidth, dheight));
  fullscreen = last_fullscreen;

  x1 = x0;
  y1 = y0;

  draw_paste_rectangle();
}

void Canvas::draw_paste_rectangle(void) {
  PPClipboard *cb = sp->getClipboard();
  QRect r0 = cb->rect();

  int x, y, w, h;
  if (zoom >= 0) {
    x = (r0.x() - shift_x) << zoom;
    y = (r0.y() - shift_y) << zoom;
    w = (r0.width() + 1) << zoom;
    h = (r0.height() + 1) << zoom;
  } else {
    int unzoom = -zoom;
    x = (r0.x() - shift_x) >> unzoom;
    y = (r0.y() - shift_y) >> unzoom;
    w = (r0.width() + 1) >> unzoom;
    h = (r0.height() + 1) >> unzoom;
  }
  if (orientation) {
    int t;
    t = x; x = y; y = dwidth - t - w;
    t = w; w = h; h = t;
  }

  if (!fullscreen) y -= parent->menuHeight();
  QRect r1(x, y, w, h);
  QPainter p(this);
  
  p.setPen(Qt::white);
  p.setRasterOp(Qt::XorROP);
  p.drawRect(r1 & rect());
}

void Canvas::set_move_mode(void) {
  if (shift_mode)
    set_shift_mode();
  else if (layer_shift_mode)
    set_layer_shift_mode();
  else
    pen_mode = PenModeMove;
}

void Canvas::set_zoom_mode(void) {
  pen_mode = PenModeZoom;
}

void Canvas::set_shift_mode(void) {
  shift_mode = !shift_mode;
  dropping_mouse_release = !shift_mode;

  sp->setToolPanelActive(ToolPanel::Move, shift_mode);
  update_panels();

  if (!shift_mode && zoom <= 0) {
    QPoint pt(shift_x, shift_y);
    sp->shiftAllLayers(pt);
    shift_x = shift_y = 0;
    repaint();
  }

  hideInfo();
}

void Canvas::set_layer_shift_mode(void) {
  layer_shift_mode = !layer_shift_mode;
  dropping_mouse_release = !layer_shift_mode;

  sp->setToolPanelActive(ToolPanel::Move, layer_shift_mode);
  update_panels();

  if (!layer_shift_mode)
    sp->fixShiftCurrentLayer();

  hideInfo();
}

void Canvas::unset_shift_mode(void) {
  shift_mode = false;
  dropping_mouse_release = !shift_mode;

  sp->setToolPanelActive(ToolPanel::Move, false);
  update_panels();

  if (zoom <= 0) {
    shift_x = shift_y = 0;
    updateRect(rect());
  }

  hideInfo();
}

void Canvas::unset_layer_shift_mode(void) {
  layer_shift_mode = false;
  dropping_mouse_release = !layer_shift_mode;

  sp->setToolPanelActive(ToolPanel::Move, false);
  update_panels();

  sp->unshiftCurrentLayer();
  updateRect(QRect(0, 0, width(), height()));

  hideInfo();
}

void Canvas::zoom_mode(void) {
  zoom_mode_flag = !zoom_mode_flag;

  sp->setToolPanelActive(ToolPanel::Zoom, zoom_mode_flag);
  update_panels();

  hideInfo();
}

void Canvas::set_zoom(int x, int y) {
  int last_panel_position = panel_position;

  hideInfo();

  if (panel_position) {
    panel_position = HideTools;
    sp->setShowingTools(false);
    update_info_rect();
  }
  
  if (zoom < 0)
    shift_x = shift_y = zoom = 0;
  else
    zoom ++;

  sp->setToolPanelActive(ToolPanel::Zoom, false);
  
  if (zoom > ZoomMax) zoom = ZoomMax;

  if (zoom)
    set_shift_by_center(x , y);

  zoom_mode_flag = false;
  repaint();

  if (last_panel_position) {
    panel_position = last_panel_position;
    sp->setShowingTools(true);
    update_info_rect();
  }
}

void Canvas::set_unzoom(void) {
  zoom --;

  if (zoom < ZoomMin) zoom = ZoomMin;

  if (zoom == 0)
    shift_x = shift_y = 0;

  adjust_shift();

  repaint();
  
  update_panels();
}

void Canvas::show_hide_menu(void) {
  static bool showing = FALSE;

  if (shift_mode) set_shift_mode();

  hideInfo();
  
  showing = !showing;
  parent->setShowMenu(showing);

  fullscreen = !showing;

  if (fullscreen) {
    updateRect(rect());
    update_panels();
  } else
    repaint();
}

void Canvas::show_hide_tools(void) {
  hideInfo();

  panel_position = (panel_position + 1) % NumOfToolsPosition;
  sp->setShowingTools(panel_position);

  update_info_rect(true);
  repaint();
}

void Canvas::show_hide_layers(void) {
  if (opt_panel_type == Layers) {
    opt_panel_type = None;
    layers.setShow(false);
  } else {
    opt_panel_type = Layers;
    layers.setShow(true);
  }

  update_info_rect();
  repaint();
}

void Canvas::set_pen_pressure_width(bool f) {
  apply_press_width = f;

  tools->setActive(ToolPanel::PencilL, f);

  update_info_rect();
}

void Canvas::set_pen_pressure_density(bool f) {
  apply_press_density = f;

  tools->setActive(ToolPanel::EraserL, f);

  update_info_rect();
}

void Canvas::hideInfo(void) {
  if (panel_position) return;
  
  if (info_rect.width()) {
    info_rect = sp->hideInfo();
    update_info_rect(true);

    info_rect.setWidth(0);
  }
}

void Canvas::load(const QString &fn) {
  sp->setOrientation(orientation);

  sp->load(fn);

  zoom = shift_x = shift_y = 0;
  img_modified = true;
}

void Canvas::loadLayer(const QString &fn) {
  sp->setOrientation(orientation);

  sp->loadLayer(fn);

  zoom = shift_x = shift_y = 0;
  img_modified = true;
}

void Canvas::save(const QString &fn, const char* ft,
                  bool warn) {
  QString fn2(fn);
  QFileInfo info(fn);

  sp->setOrientation(orientation);

  info.setFile(fn2);

  if (warn && LayeredPainter::fileExists(fn2))
    if (QMessageBox::warning(this, "PetitePeinture",
                             "File " + info.fileName() + " exists.\n" +
                             "Want to overwrite it?\n\n",
                             "Save", "Cancel"))
      return;

  if (!sp->save(fn2, ft))
    QMessageBox::warning(this, "PetitePeinture",
                         "FAILED to save the file " +
                         info.fileName() + "!\n",
                         "OK");
}

void Canvas::saveIcon(const QString &fn) {
  PtptFormat ptpt(1);

  QImage img;
  sp->getQImage(&img, -1);
  ptpt.setThumbnail(img);
  ptpt.saveThumbnail(fn);
}

void Canvas::saveLayer(const QString &fn) {
  QString fn2(fn);
  QFileInfo info(fn);

  sp->setOrientation(orientation);

  if (info.extension().isEmpty())
    fn2 += ".png";

  info.setFile(fn2);

  if (LayeredPainter::fileExists(fn2))
    if (QMessageBox::warning(this, "PetitePeinture",
                             "File " + info.fileName() + " exists.\n" +
                             "Want to overwrite?\n\n",
                             "Save", "Cancel"))
      return;

  if (!sp->saveLayer(fn2))
    QMessageBox::warning(this, "PetitePeinture",
                         "FAILED to save the file " +
                         info.fileName() + "!\n",
                         "OK");
}

void Canvas::clearImage(bool modify) {
  //  setSize(dwidth, dheight);
  setSize(0, 0);
  sp->clear();

  if (parent->noviceMode()) {
    sp->createLayer();
    sp->setCompositionMethod(1, MulComposition);
  }

  img_modified = modify;
}

void Canvas::clearCurrentLayer(void) {
  if (QMessageBox::warning(this, "PetitePeinture",
                           "Discard current layer?\n\n",
                           "Clear Layer", "Cancel"))
    return;
  
  sp->fill(sp->paperColor(), false);
  img_modified = true;
  repaint();
}

void Canvas::quit_application(void) {
  bool last_fullscreen;

  last_fullscreen = fullscreen;
  fullscreen = FALSE;
  
  if (!QMessageBox::warning(this, "PetitePeinture",
                           "Quit this application?\n\n",
                            "Quit", "Cancel")) {
    parent->quit_application();
  }

  fullscreen = last_fullscreen;
}

void Canvas::tapping(void) {
  switch (pressing_tool) {
  case 0:
    return;
    break;
  case ToolPanel::CloudL:
    if (pen != &cloud) set_cloud();
    change_pen_width(-1);
    break;
  case ToolPanel::EraserL:
    if (pen != &eraser) set_eraser();
    change_pen_width(-1);
    break;
  case ToolPanel::BrushL:
    if (pen != &brush) set_brush();
    change_pen_width(-1);
    break;
  case ToolPanel::PencilL:
    if (pen != &pencil) set_pencil();
    change_pen_width(-1);
    break;
  case ToolPanel::CloudR:
    if (pen != &cloud) set_cloud();
    change_pen_width(1);
    break;
  case ToolPanel::EraserR:
    if (pen != &eraser) set_eraser();
    change_pen_width(1);
    break;
  case ToolPanel::BrushR:
    if (pen != &brush) set_brush();
    change_pen_width(1);
    break;
  case ToolPanel::PencilR:
    if (pen != &pencil) set_pencil();
    change_pen_width(1);
    break;
  }

  timer->start(press_repeat, true);
}

void Canvas::mirrorHorizontal(void) {
  if (orientation)
    sp->mirrorVertical();
  else
    sp->mirrorHorizontal();

  sp->backupCurrentLayer();
  img_modified = true;
}

void Canvas::mirrorVertical(void) {
  if (orientation)
    sp->mirrorHorizontal();
  else
    sp->mirrorVertical();
  
  sp->backupCurrentLayer();
  img_modified = true;
}

void Canvas::rotateCW(void) {
  sp->rotateCW();

  sp->backupCurrentLayer();
  img_modified = true;
}

void Canvas::rotateCCW(void) {
  sp->rotateCCW();

  sp->backupCurrentLayer();
  img_modified = true;
}

void Canvas::scaleLayer(uint sx, uint sy) {
  if (orientation)
    sp->scaleCurrentLayer(sy, sx);
  else
    sp->scaleCurrentLayer(sx, sy);

  img_modified = true;
}

uint Canvas::canvasWidth(void) {
  if (orientation)
    return sp->get_height();
  else
    return sp->get_width();
}

uint Canvas::canvasHeight(void) {
  if (orientation)
    return sp->get_width();
  else
    return sp->get_height();
}

void Canvas::setSize(int w, int h) {
  int _w, _h;

  set_orientation();

  if (orientation) {
    _w = h & -16;
    _h = w & -16;
  } else {
    _w = w & -16;
    _h = h & -16;
  }
  _w = max(_w, dwidth);
  _h = max(_h, dheight);

  sp->setSize(_w, _h, orientation);

  pix = new QPixmap();
  sp->getQImage(&img, rect(), orientation);
  pix->convertFromImage(img, 0);
  img.detach();

  shift_x = shift_y = 0;
}

void Canvas::getPenProperties(void) {
  switch (sp->getPenMethod()) {
  case SolidBrush:
  case EraserBrush:
    set_pencil();
    break;
  case WaterBrush:
    set_brush();
    break;
  case CloudWideBrush:
    set_cloud();
    break;
  }

  PtPen sp_pen = sp->getPen();
  pen->setWidth(sp_pen.width());
  pen->setColor(sp_pen.color());

  set_pen_density(sp->getPenDensity());

  sp->setPen(*pen);
}

void Canvas::layerChanged(int) {
  if (separate_pen)
    getPenProperties();
}
