#ifndef CANVAS_H
#define CANVAS_H

#include <qmainwindow.h>
#include <qwidget.h>
#include <qvbox.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qpointarray.h>
#include <qlist.h>
#include <qcolordialog.h>
#include <qrect.h>
#include <qimage.h>
#include <qmenubar.h>

#include "layeredpainter.h"
#include "colorpanel.h"
#include "layerpanel.h"
#include "ptpen.h"
#include "z_color_adjust.h"

class SPainter;
class ToolPanel;
class QDirectPainter;
class QTimer;

class Canvas : public QWidget {
  Q_OBJECT

public:
  enum {PenPencil = 0, PenWater, PenEraser, PenCloud,
    NumOfPens};
  
  Canvas(QWidget *parent = 0);
  ~Canvas();

  void load(const QString &);
  void save(const QString &, const char *, bool = true);
  void saveIcon(const QString &);
  void loadLayer(const QString &);
  void saveLayer(const QString &);
  void applyPressureToWidth(bool f) {apply_press_width = f;}
  void applyPressureToDensity(bool f) {apply_press_density = f;}

  void setSyncWidth(bool);
  enum {LeftRightSwitchMethod = 0, LeftRightDensity,
    NumOfLeftRights};
  void setLeftRight(uint);
  void setSeparatePen(bool);
  void getPenProperties(void);
  
  void scaleLayer(uint, uint);
  uint canvasWidth(void);
  uint canvasHeight(void);
  void setSize(int, int);
  int getOrientation(void) {return orientation;};

  void adjustColorTable(bool);

public slots:
  void clearImage(bool = true);
  void clearCurrentLayer(void);
  void setPaperColor(void);
  void setUserPalette(void);
  void exchangeColorPaperPen(void);
  void show_hide_menu(void);
  void show_hide_tools(void);
  void show_hide_layers(void);

  void set_pen_pressure_width(bool);
  void set_pen_pressure_density(bool);

  void update_info_rect(void);
  void getColor(uint16);
  void finishColorPanel(void);
  void joinShowingLayers(void);
  void exchangeLayers(int, int);
  void exchangeLayersBaseColor(void);
  void exchangeLayersBaseOver(void);
  void exchangeLayersColorOver(void);
  void addNewLayer(void);
  void duplicateCurrentLayer(void);
  void deleteCurrentLayer(void);
  void moveUpCurrentLayer(void);
  void moveDownCurrentLayer(void);

  void setMinComposition(void);
  void setMaxComposition(void);
  void setMulComposition(void);
  void setScreenComposition(void);
  void setSatComposition(void);
  void setColComposition(void);
  void setNormalComposition(void);
  void setMaskComposition(void);
  void setAlphaChannelComposition(void);
  void quit_application(void);
  void set_pen_color_dlg(QColor);
  void tapping(void);

  void mirrorHorizontal(void);
  void mirrorVertical(void);
  void rotateCW(void);
  void rotateCCW(void);
  void set_select_mode_copy(void);
  void set_select_mode_paste(void);
  void set_select_mode_paste_freeze(void);
  void set_select_mode_paste_cancel(void);
  void undo(void);
  void clearClipboard(void) {sp->getClipboard()->resize(0, 0);}
  void layerChanged(int);

  void setPencil(int w) {set_pencil(); set_pen_width(w);}
  void setBrush(int w) {set_brush(); set_pen_width(w);}
  void setEraser(int w) {set_eraser(); set_pen_width(w);}
  void setCloud(int w) {set_cloud(); set_pen_width(w);}

 signals:
  void pencilWidthChanged(int);
  void brushWidthChanged(int);
  void eraserWidthChanged(int);
  void cloudWidthChanged(int);

public:
  void paintEvent(QPaintEvent *);
  void mousePressEvent(QMouseEvent *);
  void mouseMoveEvent(QMouseEvent *);
  void mouseReleaseEvent(QMouseEvent *);
  void keyPressEvent(QKeyEvent *);
  void keyReleaseEvent(QKeyEvent *);

protected:
  SPainter *parent;
  int x0, y0, x1, y1;
  int px0, py0, px1, py1;
  int min_x, min_y, max_x, max_y;
  int dwidth, dheight, xoff, yoff;
  PtPen pencil, brush, eraser, cloud, *pen;
  QList<QPoint> pts;
  QPixmap *pix;
  QImage img;
  LayeredPainter *sp;
  ToolPanel *tools;
  ColorPanel colorpanel;
  LayerPanel layers;
  bool fullscreen, pick_color_mode, zoom_mode_flag, shift_mode;
  bool layer_shift_mode, img_modified;
  enum {SelectModeNone = 0, SelectModePre, SelectModeCopy, SelectModePaste};
  uint select_mode;
  enum {PenModeNone = 0, PenModeDensity, PenModeCloud, PenModeWidth,
    PenModePressureWidth, PenModePressureDensity,
	PenModeColor, PenModeAntialias, PenModeMove, PenModeZoom,
        NumOfPenMode};
  uint pen_mode;
  int pen_width_diff;
//  bool pen_density_mode, cloud_density_mode;
//  bool pen_pressure_mode;
  enum {LayerModeNone = 0, LayerModeAlpha, LayerModeTab};
  uint layer_mode;
  QSize loupe_size;
  QRect info_rect, tools_rect, opt_panel_rect;
  QString info_string;
  int orientation;
  int zoom, shift_x, shift_y, loupe_x, loupe_y, loupe_zoom;
  enum {HideTools = 0, TopTools, BottomTools, NumOfToolsPosition};
  int panel_position;
  enum OptionalPanelType {None = 0, Tools, Color, Layers, Loupe, Num};
  OptionalPanelType opt_panel_type;
  int last_pen_method;
  int pen_density;
  int previous_value;
  bool antialias, dropping_mouse_release;
  bool apply_press_width, apply_press_density;
//  bool use_pen_pressure;
//  bool show_layers;
  bool sync_width, separate_pen;
  uint leftright;
  int pressing_tool;
  static const int press_delay = 500;
  static const int press_repeat = 50;
  QTimer *timer;
  QRect copy_rect, pcopy_rect;
  ZColorAdjust _z_col_adjust;
  static uint16 *_col_adj_table;

  void updateRect(const QRect &);
  void setupPainter(void);
  void set_antialias(bool);
  uint get_pen_method(void);
  void set_pen_method(uint);
  void set_pen_width(int);
  void set_pen_density(int);
  void set_cloud_density(int);
  void change_pen_width(int);
  void change_pen_width(void);

  void set_pen(void);
  void set_pencil(void);
  void set_brush(void);
  void set_eraser(void);
  void set_cloud(void);

  void set_pencil_mode(bool);
  void set_eraser_mode(bool);
  void set_brush_mode(bool);
  void set_cloud_mode(bool);
  void set_antialias_mode(void);

  void set_layer_alpha(int);
  void set_layer_alpha(int, int);
  void fix_layer_alpha(void);
  void set_color_panel(void);
  void set_loupe_panel(void);
  void set_peninfo_mode(void);
  void set_pick_mode(void);
  void pick_color(QMouseEvent*);
  void set_select_mode(void);
  void set_move_mode(void);
  void set_zoom_mode(void);
  void set_shift_mode(void);
  void set_layer_shift_mode(void);
  void unset_shift_mode(void);
  void unset_layer_shift_mode(void);
  void zoom_mode(void);
  void set_zoom(int, int);
  void set_unzoom(void);
  void set_shift_by_center(int, int);
  void adjust_shift(void);
  void set_loupe_by_center(int, int);
  void adjust_loupe(void);
  void shift_image(int x, int y);
  void shift_image_by_mouse(void);
  void layer_shift_image(int, int);
  void layer_shift_image_by_mouse(void);
  void select_copy_press_event(QMouseEvent*);
  void select_copy_move_event(QMouseEvent*);
  void select_copy_release_event(QMouseEvent*);
  void select_paste_press_event(QMouseEvent*);
  void select_paste_move_event(QMouseEvent*, bool = true);
  void draw_paste_rectangle(void);
  void hideInfo(void);
  void set_orientation(void);
  void update_info_rect(bool);
  void update_info_string(bool = false);
  void panel_event(QMouseEvent *e);
  void panel_move_event(QMouseEvent *e);
  void panel_release_event(QMouseEvent *e);
  void layerpanel_event(QMouseEvent *);
  void layerpanel_move_event(QMouseEvent *e);
  void layerpanel_release_event(QMouseEvent *e);
  void colorpanel_event(QMouseEvent *);
  void colorpanel_move_event(QMouseEvent *e);
  void colorpanel_release_event(QMouseEvent *e);
  void loupe_event(QMouseEvent *);
  void loupe_move_event(QMouseEvent *);
  void loupe_release_event(QMouseEvent *);

private:
  static const int PanelEventIndex = -2;
  static const int ColorPanelEventIndex = -3;
  static const int LayerPanelEventIndex = -4;

  void updateRectR0(const QRect &);
  void updateRectR270(const QRect &);
  void update_rect(const QRect &);
  void update_rect_zoom(const QRect &);
  void update_rect_nozoom(const QRect &);
  void update_rect_unzoom(const QRect &);
  void update_panel_rect(const QSize&, uint16*);
  void update_optional_panel_rect(const QSize&, uint16*);
  void update_panel_rect(void);
  void update_tools_rect(void);
  void update_colorpanel_rect(void);
  void update_layers_rect(void);
  void update_loupe_rect(void);
  void update_loupe_rect(QRect);
  void update_optional_panel(void);
  void update_panels(void);
  void update_image_info(void);
  void convert_geometry(int*, int*);
  void disconvert_geometry(int*, int*);
  QPoint convert_geometry(const QPoint&);
  void convert_loupe_geometry(int*, int*);

  inline void add_each_color(uint32 a, uint32 &r, uint32 &g, uint32 &b) {
    r += (a >> 11) & 0x001f001f;
    g += (a >> 5) & 0x003f003f;
    b += a & 0x001f001f;
  }

  inline uint32 pack_each_color(uint32 &r, uint32 &g, uint32 &b) {
    r = (r >> -zoom * 2 - 1) & 0x001f001f;
    g = (g >> -zoom * 2 - 1) & 0x003f003f;
    b = (b >> -zoom * 2 - 1) & 0x001f001f;
    r = ((r & 0xffff) + (r >> 16)) >> 1;
    g = ((g & 0xffff) + (g >> 16)) >> 1;
    b = ((b & 0xffff) + (b >> 16)) >> 1;
    return (r << 11) | (g << 5) | b;
  }

  inline uint32 adjust(uint32 s) {
    return (((uint32)_col_adj_table[s >> 16]) << 16) |
      _col_adj_table[s & 0xffff];
  }
  inline uint16 adjust(uint16 s) {
    return _col_adj_table[s];
  }

};

#endif		// CANVAS_H
