/*
 **	toolpanel.cpp
 **	by Saki Sakira <sakira@sun.dhis.portside.net>
 **	from 2003 July 4
 */

#include <stdio.h>
#include <qimage.h>
#include <cmath>
#include <qfile.h>
#include <qpe/resource.h>

#include "constants.h"
#include "sketchpainter.h"
#include "colorpalette.h"
#include "ptpen.h"
#include "toolpanel.h"

const QString ToolPanel::ToolPanelFileName = "petitpeintu_toolpanel";

ToolPanel::ToolPanel(QString &fname, int w) {
  initialize(fname, w);
}

ToolPanel::ToolPanel(int w) {
  QString fname = QString(Resource::findPixmap(ToolPanelFileName));
  
  initialize(fname, w);
}

ToolPanel::~ToolPanel() {
  delete buf;
  delete sp;
  delete color_palette;
}

void ToolPanel::initialize(QString &fname, int w) {
  QImage img;

  if (fname.isNull())
    fname = Resource::findPixmap(ToolPanelFileName);

  if (img.load(fname)) {
    QColor col;
    int x, y, fx;

    img.convertDepth(32);

    width = w > img.width() ? img.width() : w;
    width &= ~1;
    height = img.height();
    buf = new unsigned short int[width * height];
    sp = new SketchPainter(width, height);

    if (width >= img.width())
      ratio = 1.0;
    else
      ratio = ((double)img.width()) / width;

    for (y = 0; y < height; y ++)
      for (x = 0; x < width; x ++) {
        fx = (int)(x * ratio);
        if (img.valid(fx, y)) {
          col.setRgb(img.pixel(fx, y));
          buf[y * width + x] =
            SketchPainter::pack_color(col);
        } else
          buf[y * width + x] = 0xffff;
      }

    setup_tools();
  } else {
    qFatal("cannot open image %s", fname.latin1());
  }
}

void ToolPanel::setup_tools(void) {
  int w, h, x, y;

  tools.resize(NumOfTools);

  tools[Palette].setRect(0, 0,
                         (int)(460/ratio), 17);
  color_palette = new ColorPalette(
    tools[Palette].width(),
    tools[Palette].height());

  x = 1; y = 20;
  w = 44; h = 24;
  int i, j;
  for (i = Undo; i <= Pencil; i ++) {
    tools[i].setRect((int)ceil(x/ratio), y, (int)(w/ratio), h);
    if (i >= Select) {
      j = SelectL + (i - Select) * 2;
      tools[j + 0].setRect((int)ceil(x/ratio), y,
                           (int)(w/ratio)/2, h);
      tools[j + 1].setRect((int)ceil(x/ratio) + (int)(w/ratio)/2,
                           y, (int)(w/ratio)/2, h);
    }
    x += 48;
  }

  tools[PenInfo].setRect((int)ceil(432/ratio),
                         18, (int)(29/ratio), 29);

  x = (int)ceil(461/ratio);
  w = (int)(19/ratio);
  tools[Layer].setRect(x, 0, w, 48);

  for (uint i = 0; i < num_of_layers; i ++)
    tools[Layers + i].setRect(x, (num_of_layers - 1 - i) * 8,
                              w, 8);

  active.fill(false, NumOfTools);
  layers.fill(true, num_of_layers);
}


QSize ToolPanel::size(void) {
  return QSize(width, height);
}

int ToolPanel::getIndex(int x, int y) {
  for (uint i = Palette; i < NumOfTools; i ++)
    if (tools[i].contains(x, y))
      return i;

  return -1;
}

bool ToolPanel::isLeftside(int x, int y) {
  for (uint i = ZoomL; i < Layers; i += 2)
    if (tools[i].contains(x, y))
      return true;

  return false;
}

QColor ToolPanel::getColor(int x, int y) {
  return SketchPainter::unpack_color(buf[y * width + x]);
}

void ToolPanel::setActive(int i, bool f) {
  if (i >= 0 && i < (int)NumOfTools)
    active[i] = f;
}

void ToolPanel::clearSelection(void) {
  static const int infos[]
    = {Cloud, Eraser, Brush, Pencil, Select};
  static const int l = sizeof(infos) / sizeof(infos[0]);

  for (int i = 0; i < l; i ++)
    active[infos[i]] = false;
}

void ToolPanel::clearSelectionHalf(void) {
  static const int infos[] = {
    CloudL, CloudR, EraserL, EraserR,
    BrushL, BrushR, PencilL, PencilR,
    MoveL, MoveR, SelectL, SelectR};
  static const int l = sizeof(infos) / sizeof(infos[0]);

  for (int i = 0; i < l; i ++)
    active[infos[i]] = false;
}

void ToolPanel::setPen(const PtPen &p) {
  pen = p;
  sp->setPen(p);
}

void ToolPanel::setPenDensity(int d) {
  density = d;
  sp->setPenDensity(d);
}

void ToolPanel::setColorPalette(const QString &fn) {
  uint w, h, x0, y0, x, y;
  uint16 *pbuf;

  color_palette->setImage(fn);

  w = tools[Palette].width();
  h = tools[Palette].height();
  x0 = tools[Palette].left();
  y0 = tools[Palette].top();
  pbuf = new uint16[w * h];

  color_palette->infoImage(pbuf);

  for (y = 0; y < h; y ++)
    for (x = 0; x < w; x ++)
      buf[(y0 + y) * width + x0 + x]
        = pbuf[y * w + x];

  delete pbuf;
}

void ToolPanel::initPalette(void) {
  QString fn = Resource::findPixmap(ToolPanelFileName);
  initialize(fn, width);
}

void ToolPanel::copy_buf(void) {
  unsigned long int *buf32 = (unsigned long int*)buf;
  unsigned long int *spbuf32 =
    (unsigned long int*)sp->frameBuffer();
  int l = width * height >> 1;

  for (int i = 0; i < l; i ++)
    spbuf32[i] = buf32[i];
}

void ToolPanel::setPenMethod(int p) {
  sp->setPenMethod(p);
}

void ToolPanel::setShowLayer(int i, bool f) {
  if (i < (int)num_of_layers)
    layers[i] = f;
}

unsigned short int* ToolPanel::infoImage(void) {
  copy_buf();
  
  const uint16 white = 0xffff;
  const uint16 yellow = 0xffe0;

  for (uint i = 0; i < num_of_layers; i ++)
    if (layers[i])
      sp->fillRect16(tools[Layers + i], yellow);
    else
      sp->fillRect16(tools[Layers + i], white);
  
  for (uint i = 0; i < NumOfTools; i ++)
    if (active[i])
      sp->invertRect(tools[i]);

  int cx, cy;
  QRect *pi = &tools[PenInfo];
  sp->fillRect32(*pi, 0xffff0000);
  cx = pi->x() + (pi->width() >> 1);
  cy = pi->y() + (pi->height() >> 1);

//  sp->setPenMethod(active[Brush] ? WaterBrush : SolidBrush);
  sp->clearMask();
//  sp->setPenDensity(density);
  sp->drawPoint(cx, cy);
  sp->clearMask();

  if (active[PenInfo]) {
    QRect rect = tools[PenInfo];

    rect.setHeight(rect.height() / 2);
    sp->invertRect(rect);
  }

  return sp->frameBuffer();
}
