/*
 *	tspressure.cpp
 *	touchscreen pressure reader
 *	by Saki Sakira <sakira@sun.dhis.portside.net>
 *	from 2003 Aug 19
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#include <qdialog.h>
#include <qvbox.h>
#include <qwidget.h>
#include <qhbox.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qcheckbox.h>
#include <qgrid.h>

#include "tspressure.h"

////////////////////////////////////////////////////////////
// TSPressure

const char* TSPressure::DevFN = "/dev/sharp_ts";

unsigned long int TSPressure::pressure_min = 0;
unsigned long int TSPressure::pressure_mid = 0;
unsigned long int TSPressure::pressure_max = 0;
unsigned long int TSPressure::pressure_sum = 0;
int TSPressure::smooth = 4;
QArray<uint32> TSPressure::last_press;
float TSPressure::a = 0.0;
float TSPressure::b = 0.0;
TSPressure::press_dist TSPressure::p_max;
TSPressure::press_dist TSPressure::p_min;
uint TSPressure::width = 0;
uint TSPressure::height = 0;

TSPressure::TSPressure(void) {
  last_press.fill(pressure_min, smooth);

  setCoefficients();
}

int TSPressure::pressure(void) {
  long int p = rarePressure();
  if (p == (long int)PressureError) return -1;

  return pressure(p);
}

int TSPressure::pressure(int px, int py) {
  long int press = rarePressure();
  
  if (press == (long int)PressureError) return -1;

  int press_min = min_press(px, py);
  int press_max = max_press(px, py);

  //  printf("%d %d %d\n", press_min, press_max, (int)press);
  
  long int p = pressure_min;

  if (press_max != press_min)
    p += (press - press_min) * (pressure_max - pressure_min)
      / (press_max - press_min);

  return pressure(p);
}

int TSPressure::pressure(long int p) {
  static int index;
  static int r;
  float x;

  if (p > (long int)(pressure_max + pressure_min)) {
    return r;
  } else {

    if (p > (long int)pressure_max) p = pressure_max;
    if (p < (long int)pressure_min) p = pressure_min;

    pressure_sum -= last_press[index];
    pressure_sum += p;

    last_press[index ++] = p;
    if (index >= smooth) index = 0;


    p = pressure_sum / smooth;

    x = (float)(p - pressure_min);

    r = (int)(a*x*x + b*x);

    if (r < 0)
      r = 0;
    else if (r > 256)
      r = 256;
  }

  return r;
}

void TSPressure::clear(void) {
  last_press.fill(pressure_min, smooth);
  pressure_sum = pressure_min * smooth;
}

inline unsigned long int TSPressure::rarePressure(void) {
  int fd;
  unsigned long int p;

  fd = open(DevFN, O_WRONLY);
  if (fd == -1 || ioctl(fd, ReadPressure, &p)) {
    if (fd != -1)
      close(fd);
    return PressureError;
  }
  close(fd);

  return p;
}

unsigned long int TSPressure::pressureMin(void) {
  return pressure_min;
}

unsigned long int TSPressure::pressureMid(void) {
  return pressure_mid;
}

unsigned long int TSPressure::pressureMax(void) {
  return pressure_max;
}

void TSPressure::setPressureMin(unsigned long int i) {
  pressure_min = i;
  setCoefficients();
}

void TSPressure::setPressureMid(unsigned long int i) {
  pressure_mid = i;
  setCoefficients();
}

void TSPressure::setPressureMin(press_dist p) {
  p_min = p;
}

void TSPressure::setPressureMax(press_dist p) {
  p_max = p;
}

void TSPressure::setPressureMax(unsigned long int i) {
  pressure_max = i;
  setCoefficients();
}

void TSPressure::setCoefficients(void) {
  double den, d, e;

  d = (double)(pressure_mid - pressure_min);
  e = (double)(pressure_max - pressure_min);

  if (d <= 0) d = 1;
  if (e <= d) e = d + 1;

  /*
    ad^2 + bd = 128,
    ae^2 + be = 256.
   */

  den = e * d * (e - d);

  a = (float)((256*d - 128*e) / den);
  b = (float)(128 * (-2*d*d + e*e) / den);
}


int TSPressure::getSmooth(void) {
  return smooth;
}

void TSPressure::setSmooth(int s) {
  if (s < 1)
    s = 1;
  else if (s > 100)
    s = 100;

  smooth = s;
  last_press.resize(s);

  clear();
}

int TSPressure::min_max_press(int x, int y, uint w, uint h,
			      int p00, int p10, int p01) {
  long long int p_0y;
  
  p_0y = p00 + (p01 - p00) * (long long int)(y * y) / (h * h);
  return p_0y + (p10 - p_0y) * (long long int)(x * x) / (w * w);
}

int TSPressure::min_max_press(int x, int y, const press_dist &pd) {
  uint w = width >> 1;
  uint h = height >> 1;

  if (w && h && pd.oo && pd.mo && pd.om && pd.po && pd.op) {
    if (y < (int)h) {
      if (x < (int)w)
	return min_max_press(w - x, h - y, w, h, pd.oo, pd.mo, pd.om);
      else
	return min_max_press(x - w, h - y, w, h, pd.oo, pd.po, pd.om);
    } else {
      if (x < (int)w)
	return min_max_press(w - x, y - h, w, h, pd.oo, pd.mo, pd.op);
      else
	return min_max_press(x - w, h - y, w, h, pd.oo, pd.po, pd.op);
    }
  } else {
    if (&pd == &p_max)
      return pressure_max;
    else
      return pressure_min;
  }
}

int TSPressure::min_press(int x, int y) {
  return min_max_press(x, y, p_min);
}

int TSPressure::max_press(int x, int y) {
  return min_max_press(x, y, p_max);
}

////////////////////////////////////////////////////////////
// PressureSensor

PressureSensor::PressureSensor(QWidget* w) : QWidget(w) {
}

void PressureSensor::mouseMoveEvent(QMouseEvent *) {
  pressure = TSPressure::rarePressure();

  emit pressed(pressure);
}


////////////////////////////////////////////////////////////
// TSDialog

#include <qapplication.h>
#include <qpushbutton.h>
#include <qsizepolicy.h>

unsigned long int TSDialog::min_pressure = 0;
unsigned long int TSDialog::mid_pressure = 0;
unsigned long int TSDialog::max_pressure = 0;

TSDialog::TSDialog(void) : QDialog(0, "TSDialog", true) {
  setCaption("Touchscreen Config.");

  //  setFixedSize(qApp->desktop()->size());

  main_box = new QVBox(this);
  main_box->setMinimumSize(240, 300);
  //  main_box->setSizePolicy(QSizePolicy(QSizePolicy::Expanding,
  //				      QSizePolicy::Expanding));
  canvas = new PressureSensor(main_box);
  canvas->setMinimumSize(160, 50);
  canvas->setBackgroundColor(Qt::white);
  labels_box = new QHBox(main_box);
  nums_box = new QHBox(main_box);
  l_min = new QLabel("min.", labels_box);
  l_mid = new QLabel("mid.", labels_box);
  l_max = new QLabel("max.", labels_box);
  e_min = new QLineEdit("", nums_box);
  e_mid = new QLineEdit("", nums_box);
  e_max = new QLineEdit("", nums_box);
  smooth_box = new QHBox(main_box);
  l_smooth = new QLabel("smooth ", smooth_box);
  e_smooth = new QLineEdit("", smooth_box);

  supple_grid = new QGrid(3, main_box);
  east_btn = new QPushButton("east", supple_grid);
  east_min = new QLineEdit("", supple_grid);
  east_max = new QLineEdit("", supple_grid);
  west_btn = new QPushButton("west", supple_grid);
  west_min = new QLineEdit("", supple_grid);
  west_max = new QLineEdit("", supple_grid);
  south_btn = new QPushButton("south", supple_grid);
  south_min = new QLineEdit("", supple_grid);
  south_max = new QLineEdit("", supple_grid);
  north_btn = new QPushButton("north", supple_grid);
  north_min = new QLineEdit("", supple_grid);
  north_max = new QLineEdit("", supple_grid);
  news = NEWSNone;

  show_values();

  connect(canvas, SIGNAL(pressed(unsigned long int)),
          this, SLOT(pressureChanged(unsigned long int)));
  connect(e_min, SIGNAL(textChanged(const QString &)),
          this, SLOT(min_pressure_changed(const QString &)));
  connect(e_max, SIGNAL(textChanged(const QString &)),
          this, SLOT(max_pressure_changed(const QString &)));
  connect(e_mid, SIGNAL(textChanged(const QString &)),
          this, SLOT(mid_pressure_changed(const QString &)));
  connect(east_btn, SIGNAL(clicked()), this, SLOT(start_east()));
  connect(west_btn, SIGNAL(clicked()), this, SLOT(start_west()));
  connect(south_btn, SIGNAL(clicked()), this, SLOT(start_south()));
  connect(north_btn, SIGNAL(clicked()), this, SLOT(start_north()));
  //  connect(this, SIGNAL(pressed(unsinged long int)),
  //	  this, SLOT(newsPressureChanged(unsigned long int)));
}

TSDialog::~TSDialog(void) {
  delete east_btn;
  delete west_btn;
  delete south_btn;
  delete north_btn;
  delete east_min;
  delete west_min;
  delete south_min;
  delete north_min;
  delete east_max;
  delete west_max;
  delete south_max;
  delete north_max;
  delete e_max;
  delete l_max;
  delete e_min;
  delete l_min;
  delete supple_grid;
  delete nums_box;
  delete canvas;
  delete main_box;
}

void TSDialog::show_values(void) {
  e_min->setText(QString::number(min_pressure));
  e_mid->setText(QString::number(mid_pressure));
  e_max->setText(QString::number(max_pressure));
}

void TSDialog::pressureChanged(unsigned long int press) {
  if (press == 0) return;
  
  if (min_pressure == 0)
    min_pressure = press;
  if (max_pressure == 0)
    max_pressure = press;

  if (press < min_pressure)
    min_pressure = press;
  if (max_pressure < press)
    max_pressure = press;

  mid_pressure = (min_pressure + 3 * max_pressure) / 4;

  show_values();
}

void TSDialog::newsPressureChanged(unsigned long int press) {
  if (press == 0) return;

  QLineEdit *_min, *_max;
  unsigned long int min_press, max_press;

  switch (news) {
  case NEWSEast:
    _min = east_min; _max = east_max;
    break;
  case NEWSWest:
    _min = west_min; _max = west_max;
    break;
  case NEWSSouth:
    _min = south_min; _max = south_max;
    break;
  case NEWSNorth:
    _min = north_min, _max = north_max;
    break;
  default:
    return;
  }

  min_press = _min->text().toULong();
  max_press = _max->text().toULong();

  if (min_press == 0) min_press = press;
  if (max_press == 0) max_press = press;

  if (press < min_press) min_press = press;
  if (max_press < press) max_press = press;

  _min->setText(QString::number(min_press));
  _max->setText(QString::number(max_press));
}

void TSDialog::min_pressure_changed(const QString& str) {
  min_pressure = str.toULong();
}

void TSDialog::mid_pressure_changed(const QString& str) {
  mid_pressure = str.toULong();
}

void TSDialog::max_pressure_changed(const QString& str) {
  max_pressure = str.toULong();
}

void TSDialog::button_pressed(int) {
  show_values();
}

int TSDialog::smooth(void) {
  return e_smooth->text().toInt();
}

void TSDialog::setSmooth(int s) {
  if (s < 1)
    s = 1;
  
  e_smooth->setText(QString::number(s));
}

void TSDialog::mousePressEvent(QMouseEvent *) {
  newsPressureChanged(TSPressure::rarePressure());
}

void TSDialog::mouseReleaseEvent(QMouseEvent *) {
  news = NEWSNone;
  releaseMouse();
}

unsigned long int TSDialog::minPressure(NEWS i) {
  switch (i) {
  case NEWSEast:
    return east_min->text().toULong();
  case NEWSWest:
    return west_min->text().toULong();
  case NEWSSouth:
    return south_min->text().toULong();
  case NEWSNorth:
    return north_min->text().toULong();
  default:
    return 0;
  }
}

unsigned long int TSDialog::maxPressure(NEWS i) {
  switch (i) {
  case NEWSEast:
    return east_max->text().toULong();
  case NEWSWest:
    return west_max->text().toULong();
  case NEWSSouth:
    return south_max->text().toULong();
  case NEWSNorth:
    return north_max->text().toULong();
  default:
    return 0;
  }
}

void TSDialog::setMinPressure(NEWS i, unsigned long int p) {
  switch (i) {
  case NEWSEast:
    east_min->setText(QString::number(p)); break;
  case NEWSWest:
    west_min->setText(QString::number(p)); break;
  case NEWSSouth:
    south_min->setText(QString::number(p)); break;
  case NEWSNorth:
    north_min->setText(QString::number(p)); break;
  default:
    return;
  }
}

void TSDialog::setMaxPressure(NEWS i, unsigned long int p) {
  switch (i) {
  case NEWSEast:
    east_max->setText(QString::number(p)); break;
  case NEWSWest:
    west_max->setText(QString::number(p)); break;
  case NEWSSouth:
    south_max->setText(QString::number(p)); break;
  case NEWSNorth:
    north_max->setText(QString::number(p)); break;
  default:
    return;
  }
}
